#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from numba import njit, prange
from numpy import any, all, sum, ndarray, zeros, array, exp, log, atleast_2d, ones, arange, atleast_1d, isnan, inf, unique

from ...orbits.orbits_py import z_ip_s, z_ip_v


MODE_NORMAL = 0
MODE_TRSPEC = 1

# Gamma and log gamma
# -------------------

lanczos_g = 6.024680040776729583740234375
lanczos_g_minus_half = 5.524680040776729583740234375

lanczos_num_coeffs = array([
    23531376880.410759688572007674451636754734846804940,
    42919803642.649098768957899047001988850926355848959,
    35711959237.355668049440185451547166705960488635843,
    17921034426.037209699919755754458931112671403265390,
    6039542586.3520280050642916443072979210699388420708,
    1439720407.3117216736632230727949123939715485786772,
    248874557.86205415651146038641322942321632125127801,
    31426415.585400194380614231628318205362874684987640,
    2876370.6289353724412254090516208496135991145378768,
    186056.26539522349504029498971604569928220784236328,
    8071.6720023658162106380029022722506138218516325024,
    210.82427775157934587250973392071336271166969580291,
    2.5066282746310002701649081771338373386264310793408])

lanczos_den_coeffs = array([
    0.0, 39916800.0, 120543840.0, 150917976.0, 105258076.0, 45995730.0,
    13339535.0, 2637558.0, 357423.0, 32670.0, 1925.0, 66.0, 1.0])

lanczos_n = lanczos_num_coeffs.size

gamma_integral = array([
    1.0, 1.0, 2.0, 6.0, 24.0, 120.0, 720.0, 5040.0, 40320.0, 362880.0,
    3628800.0, 39916800.0, 479001600.0, 6227020800.0, 87178291200.0,
    1307674368000.0, 20922789888000.0, 355687428096000.0,
    6402373705728000.0, 121645100408832000.0, 2432902008176640000.0,
    51090942171709440000.0, 1124000727777607680000.0])


@njit
def lanczos_sum(x):
    num, den = 0.0, 0.0
    if x < 0.0:
        raise ValueError
    if (x < 5.0):
        for i in range(lanczos_n - 1, -1, -1):
            num = num * x + lanczos_num_coeffs[i]
            den = den * x + lanczos_den_coeffs[i]
    else:
        for i in range(lanczos_n):
            num = num / x + lanczos_num_coeffs[i]
            den = den / x + lanczos_den_coeffs[i]
    return num / den


@njit
def loggamma(x):
    """Unsafe Log Gamma function for positive x.

    Notes
    -----
    The code is modified from the implementation in the python
    math library, but with all the safety checking removed.

    """
    absx = abs(x)
    if (absx < 1e-20):
        return -log(absx)

    r = log(lanczos_sum(absx)) - lanczos_g
    r += (absx - 0.5) * (log(absx + lanczos_g - 0.5) - 1)
    return r


@njit
def gamma(x):
    """Unsafe Gamma function for positive x.

    Notes
    -----
    The code is modified from the implementation in the python
    math library, but with all the safety checking removed.
    """
    return exp(loggamma(x))


def jcoeff(alpha, beta, ri):
    res = zeros(4)
    res[0] = 2. * ri * (ri + alpha + beta) * (2. * ri - 2. + alpha + beta)
    res[1] = (2. * ri - 1. + alpha + beta) * (2. * ri + alpha + beta) * (2. * ri - 2. + alpha + beta)
    res[2] = (2. * ri - 1. + alpha + beta) * (alpha + beta) * (alpha - beta)
    res[3] = - 2. * (ri - 1. + alpha) * (ri - 1. + beta) * (2. * ri + alpha + beta)
    return res


def init_arrays(npol, nldc):
    anm, avl = zeros((npol, nldc + 1)), zeros((npol, nldc + 1))
    ajd, aje = zeros((4, npol, nldc + 1)), zeros((4, npol, nldc + 1))

    for j in range(nldc + 1):
        nu = (j + 2.0) / 2.0
        for i in range(npol):
            nm1 = exp(loggamma(nu + i + 1.0) - loggamma(i + 2.0))
            avl[i, j] = (-1) ** (i) * (2.0 + 2.0 * i + nu) * nm1
            anm[i, j] = exp(loggamma(i + 1.0) + loggamma(nu + 1.0) - loggamma(i + 1.0 + nu))

    for j in range(nldc + 1):
        nu = (2.0 + j) / 2.0
        for i in range(npol):
            ajd[:, i, j] = jcoeff(0.0, 1.0 + nu, i + 1)
            aje[:, i, j] = jcoeff(nu, 1.0, i + 1)

    return anm, avl, ajd, aje


@njit
def jacobi(npol, alpha, beta, x, i_ld, j_c):
    """Jacobi polynomial

    Notes
    -----

    Adapted from the Jacobi polynomials routine by J. Burkardt. The
    only major difference is that the routine computes the values for
    multiple x at the same time.
    """
    j = i_ld
    cx = zeros((x.size, npol))
    cx[:, 0] = 1.0
    cx[:, 1] = (1.0 + 0.5 * (alpha + beta)) * x + 0.5 * (alpha - beta)
    for i in range(1, npol - 1):
        for k in range(x.size):
            cx[k, i + 1] = ((j_c[2, i, j] + j_c[1, i, j] * x[k]) * cx[k, i] + j_c[3, i, j] * cx[k, i - 1]) / j_c[
                0, i, j]
    return cx


@njit
def jacobi_s(npol, alpha, beta, x, i_ld, j_c):
    """Jacobi polynomial

    Notes
    -----

    Adapted from the Jacobi polynomials routine by J. Burkardt. The
    only major difference is that the routine computes the values for
    multiple x at the same time.
    """
    j = i_ld
    cx = zeros(npol)
    cx[0] = 1.0
    cx[1] = (1.0 + 0.5 * (alpha + beta)) * x + 0.5 * (alpha - beta)
    for i in range(1, npol - 1):
        cx[i + 1] = ((j_c[2, i, j] + j_c[1, i, j] * x) * cx[i] + j_c[3, i, j] * cx[i - 1]) / j_c[0, i, j]
    return cx


@njit
def alpha(b: float, c: ndarray, n: int, npol: int, nldc: int, anm: ndarray, avl: ndarray, ajd: ndarray, aje: ndarray):
    sm = zeros(c.size)
    nu = (2.0 + n) / 2.0
    norm = b * b * (1.0 - c * c) ** (1.0 + nu) / (nu * gamma(1.0 + nu))

    e = jacobi_s(npol, nu, 1.0, 1.0 - 2 * (1.0 - b), n, aje)
    for i in range(npol):
        e[i] = (e[i] * anm[i, n]) ** 2

    d = jacobi(npol, 0.0, 1.0 + nu, 1.0 - 2 * c ** 2, n, ajd)
    for i in range(npol):
        sm += avl[i, n] * d[:, i] * e[i]

    a = norm * sm
    return a


@njit
def alpha_s(b: float, c: float, n: int, npol: int, nldc: int, anm: ndarray, avl: ndarray, ajd: ndarray, aje: ndarray):
    sm = 0.0
    nu = (2.0 + n) / 2.0
    norm = b * b * (1.0 - c * c) ** (1.0 + nu) / (nu * gamma(1.0 + nu))

    e = jacobi_s(npol, nu, 1.0, 1.0 - 2 * (1.0 - b), n, aje)
    for i in range(npol):
        e[i] = (e[i] * anm[i, n]) ** 2

    d = jacobi_s(npol, 0.0, 1.0 + nu, 1.0 - 2 * c ** 2, n, ajd)
    for i in range(npol):
        sm += avl[i, n] * d[i] * e[i]

    a = norm * sm
    return a


@njit
def general_model_vz(z: ndarray, k: float, u, npol, nldc, npb, anm, avl, ajd, aje):
    u = atleast_2d(u)
    assert u.shape == (npb, nldc)
    b = k / (1.0 + k)
    c = z / (1.0 + k)

    zmask = c <= 1.0
    nz = zmask.sum()

    a = zeros((nz, nldc + 1))
    for j in range(nldc + 1):
        a[:, j] = alpha(b, c[zmask], j, npol, nldc, anm, avl, ajd, aje)

    cn = ones((npb, nldc + 1))
    n = arange(nldc + 1)
    for ui in range(npb):
        cn[ui, 0] = (1. - u[ui].sum()) / (1.0 - sum(u[ui] * n[1:] / (n[1:] + 2.0)))
        cn[ui, 1:] = u[ui] / (1.0 - sum(n[1:] * u[ui] / (n[1:] + 2.0)))

    model = zeros((npb, nz))
    for i in range(nz):
        for j in range(npb):
            model[j, i] = -sum(a[i, :] * cn[j], 0)

    flux = ones((npb, z.size))
    flux[:, zmask] += model
    return flux


@njit
def general_model_z(z, k, u, npol, nldc, npb, anm, avl, ajd, aje):
    u = atleast_2d(u)
    assert u.shape == (npb, nldc)

    b = k / (1.0 + k)
    c = z / (1.0 + k)

    if c > 1.0:
        return ones(npb)
    else:
        a = zeros(nldc + 1)
        for j in range(nldc + 1):
            a[j] = alpha_s(b, c, j, npol, nldc, anm, avl, ajd, aje)

        cn = ones((npb, nldc + 1))
        n = arange(nldc + 1)
        for ui in range(npb):
            cn[ui, 0] = (1. - u[ui].sum()) / (1.0 - sum(u[ui] * n[1:] / (n[1:] + 2.0)))
            cn[ui, 1:] = u[ui] / (1.0 - sum(n[1:] * u[ui] / (n[1:] + 2.0)))

        flux = zeros(npb)
        for j in range(npb):
            flux[j] = -sum(a * cn[j], 0)
        return flux + 1


@njit(parallel=True)
def general_model_pz(z, k, u, npol, nldc, npb, anm, avl, ajd, aje):
    flux = zeros((npb, z.size))

    for i in prange(z.size):
        flux[:, i] = general_model_z(z[i], k, u, npol, nldc, npb, anm, avl, ajd, aje)
    return flux


@njit(parallel=True, fastmath=True)
def general_model_v(t: ndarray, k: ndarray, t0: ndarray, p: ndarray, a: ndarray, i: ndarray, e: ndarray, w: ndarray, ldc: ndarray,
                    mode: int, lcids: ndarray, pbids: ndarray, nsamples: ndarray, exptimes: ndarray,
                    npb: int, npol: int, nldc: int, es: ndarray, ms: ndarray, tae: ndarray,
                    anm: ndarray, avl: ndarray, ajd: ndarray, aje: ndarray):

    k = atleast_2d(k)
    t0, p, a, i, e, w = atleast_1d(t0), atleast_1d(p), atleast_1d(a), atleast_1d(i), atleast_1d(e), atleast_1d(w)

    ldc = atleast_2d(ldc)
    if ldc.shape[1] != nldc*npb:
        raise ValueError("The quadratic model needs two limb darkening coefficients per passband")

    npv = k.shape[0]
    npt = t.size

    if mode == MODE_NORMAL:
        flux = zeros((npv, 1, npt))
        npc = 1
    elif mode == MODE_TRSPEC:
        flux = zeros((npv, npb, npt))
        npc = npb
    else:
        raise NotImplementedError

    ulcs = unique(lcids)
    nlcs = ulcs.size
    for ipv in prange(npv):

        fpv = zeros((flux.shape[1], flux.shape[2]))

        if isnan(a[ipv]) or isnan(i[ipv]):
            flux[ipv, :, :] = inf
            continue

        for ilc in range(nlcs):
            ipb = pbids[ilc]

            if k.shape[1] == 1:
                _k = k[ipv, 0]
            else:
                _k = k[ipv, ipb]

            if mode == MODE_NORMAL:
                lldc = ldc[ipv, nldc * ipb: nldc * (ipb + 1)].reshape((1, nldc))
            else:
                lldc = ldc[ipv, :].reshape((npb, nldc))

            msk = lcids == ilc

            tlc = t[msk]
            for isample in range(1, nsamples[ilc] + 1):
                time_offset = exptimes[ilc] * ((isample - 0.5) / nsamples[ilc] - 0.5)
                z = z_ip_v(tlc + time_offset, t0[ipv], p[ipv], a[ipv], i[ipv], e[ipv], w[ipv], es, ms, tae)
                fpv[:, msk] += general_model_vz(z, _k, lldc, npol, nldc, npc, anm, avl, ajd, aje)
            fpv[:, msk] /= nsamples[ilc]
        flux[ipv, :, :] = fpv

    return flux



@njit(parallel=True, fastmath=False)
def general_model_s(t: ndarray, k: ndarray, t0: float, p: float, a: float, i: float, e: float, w: float, ldc: ndarray,
                    mode: int, lcids: ndarray, pbids: ndarray, nsamples: ndarray, exptimes: ndarray,
                    npb: int, npol: int, nldc: int, es: ndarray, ms: ndarray, tae: ndarray,
                    anm: ndarray, avl: ndarray, ajd: ndarray, aje: ndarray):
    """

    Parameters
    ----------
    t: ndarray
        Mid-exposure times
    k: ndarray
        Planet-star radius ratio(s)
    t0: float
        Zero epoch
    p: float
        Orbital period
    a: float
        Semi-major axis divided by the stellar radius
    i: float
        inclination [rad]
    e: float
        Eccentricity
    w: float
        Argument of peri-astron [rad]
    ldc: ndarray
        Limb darkening coefficients, should have a size of npb*nldc.
    mode: int
        Either `0` for normal behaviour or `1` for transmission spectroscopy. The transmission spectroscopy mode
        ignores the passband indices (`pbids`) and evaluates the model over all the passbands for all the exposures.
        The evaluation is done using optimizations presented in Parviainen (2015) and is much faster than evaluating
        the model in normal mode for transmission spectroscopy. However, the approach assumes that the photometry
        has been created from a spectroscopic time series so that all the passbands are observed simultaneously.
    lcids: ndarray
        Light curve indices
    pbids: ndarray
        Passband indices
    nsamples: ndarray
        Number of samples per exposure
    exptimes: ndarray
        Exposure times
    npb: int
        Number of passbands
    npol: int
        Number of polynomials
    nldc: int
        Number of limb darkening coefficients
    es
    ms
    tae
    anm
    avl
    ajd
    aje

    Returns
    -------

    """
    ldc = atleast_1d(ldc)
    k = atleast_1d(k)
    npt = t.size

    if mode == MODE_NORMAL:
        flux = zeros((1, npt))
        npc = 1
    elif mode == MODE_TRSPEC:
        flux = zeros((npb, npt))
        npc = npb
    else:
        raise NotImplementedError

    ulcs = unique(lcids)
    nlcs = ulcs.size

    if any(isnan(k)) or isnan(a) or isnan(i):
        flux[:, :] = inf
        return flux

    for ilc in prange(nlcs):
        ipb = pbids[ilc]

        if k.size == 1:
            _k = k[0]
        else:
            _k = k[ipb]

        if mode == MODE_NORMAL:
            lldc = ldc[nldc * ipb: nldc * (ipb + 1)].reshape((1, nldc))
        else:
            lldc = ldc.reshape((npb, nldc))

        msk = lcids == ilc

        tlc = t[msk]
        for isample in range(1, nsamples[ilc] + 1):
            time_offset = exptimes[ilc] * ((isample - 0.5) / nsamples[ilc] - 0.5)
            z = z_ip_v(tlc + time_offset, t0, p, a, i, e, w, es, ms, tae)
            flux[:, msk] += general_model_vz(z, _k, lldc, npol, nldc, npc, anm, avl, ajd, aje)
        flux[:, msk] /= nsamples[ilc]
    return flux

@njit(parallel=True, fastmath=False)
def general_model_pv(t: ndarray, pvp: ndarray, ldc: ndarray,
                    mode: int, lcids: ndarray, pbids: ndarray, nsamples: ndarray, exptimes: ndarray,
                    npb: int, npol: int, nldc: int, es: ndarray, ms: ndarray, tae: ndarray,
                    anm: ndarray, avl: ndarray, ajd: ndarray, aje: ndarray):

    pvp = atleast_2d(pvp)
    nk = pvp.shape[1] - 6
    return general_model_v(t, pvp[:,:nk], t0=pvp[:,nk], p=pvp[:,nk+1], a=pvp[:,nk+2], i=pvp[:,nk+3], e=pvp[:,nk+4], w=pvp[:,nk+5],
                        ldc=ldc, mode=mode, lcids=lcids, pbids=pbids, nsamples=nsamples, exptimes=exptimes,
                        npb=npb, npol=npol, nldc=nldc, es=es, ms=ms, tae=tae, anm=anm, avl=avl, ajd=ajd, aje=aje)


@njit(parallel=True, fastmath=True)
def general_model_s_slow(t, k, t0, p, a, i, e, w, ldc, lcids, pbids, nsamples, exptimes, npb, es, ms, tae,
                    npol, nldc, anm, avl, ajd, aje):
    ldc = atleast_1d(ldc)
    k = atleast_1d(k)

    npt = t.size
    flux = zeros(npt)
    for j in prange(npt):
        ilc = lcids[j]
        ipb = pbids[ilc]

        if k.size == 1:
            _k = k[0]
        else:
            _k = k[ipb]

        if isnan(_k) or isnan(a) or isnan(i):
            flux[j] = inf
        else:
            for isample in range(1, nsamples[ilc] + 1):
                time_offset = exptimes[ilc] * ((isample - 0.5) / nsamples[ilc] - 0.5)
                z = z_ip_s(t[j] + time_offset, t0, p, a, i, e, w, es, ms, tae)
                if z > 1.0 + _k:
                    flux[j] += 1.
                else:
                    flux[j] += general_model_z(z, _k, ldc[nldc * ipb: nldc * (ipb + 1)], npol, nldc, npb, anm, avl, ajd, aje)[0]
            flux[j] /= nsamples[ilc]
    return flux



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import matplotlib as mpl
import matplotlib.pyplot as pl
import seaborn as sb
from matplotlib.gridspec import GridSpec

from numpy import linspace, median, argmin, percentile

color = sb.color_palette()[0]
color_rgb = mpl.colors.colorConverter.to_rgb(color)
colors = [sb.utils.set_hls_values(color_rgb, l=l) for l in linspace(1, 0, 12)]
cmap = sb.blend_palette(colors, as_cmap=True)

color = sb.color_palette()[1]
color_rgb = mpl.colors.colorConverter.to_rgb(color)
colors = [sb.utils.set_hls_values(color_rgb, l=l) for l in linspace(1, 0, 12)]
cmap2 = sb.blend_palette(colors, as_cmap=True)

## Color definitions
## -----------------
c_ob = "#002147"  # Oxford blue
c_bo = "#CC5500"  # Burnt orange


def plot_kdist(samples, side: str = 'right', ax=None, clip: tuple = (0, 1), percentiles: tuple = (16, 84),
               offset: float = 0.02, bw=0.005, gridsize: int = 200):
    assert side in ('left', 'right')
    sign = 1 if side == 'right' else -1
    fig, ax = (None, ax) if ax is not None else pl.subplots()
    p = sb.kdeplot(samples, kernel='cos', bw=bw, gridsize=gridsize, cut=0, clip=clip, vertical=True, ax=ax, color='k',
                   legend=False)
    xd, yd = p.lines[-1].get_xdata(), p.lines[-1].get_ydata()
    m = median(samples)
    my = xd[argmin(abs(yd - m))] / xd.max()
    p.lines[-1].set_xdata(sign * (offset + xd / xd.max()))
    ax.plot((sign * offset, sign * offset), clip, 'k')
    ax.plot((sign * offset, sign * my), (m, m), 'k')
    p = percentile(samples, percentiles)
    mask = (yd > p[0]) & (yd < p[1])
    ax.fill_betweenx(yd[mask], sign * (offset + xd[mask] / xd.max()),  sign*offset, alpha=0.25)
    return fig

def plot_two_sided_kde(left, right, clip: tuple = (0, 1), percentiles: tuple = (16, 84),
                       offset: float = 0.02, bw=0.005, gridsize: int = 200, ax = None):
    fig, ax = (None, ax) if ax is not None else pl.subplots()
    plot_kdist(left, side='left', clip=clip, percentiles=percentiles, offset=offset, bw=bw, gridsize=gridsize, ax=ax)
    plot_kdist(right, side='right', clip=clip, percentiles=percentiles, offset=offset, bw=bw, gridsize=gridsize, ax=ax)
    pl.setp(ax, xlim=(-1.1, 1.1))
    return fig

def _jplot(hte, cte, cnr, imp, rho, fw=10, nb=30, gs=25, simulation=False, **kwargs):
    htelim = kwargs.get('htelim', (2000, 8000))
    ctelim = kwargs.get('ctelim', (4000, 12000))
    blim = kwargs.get('blim', (0, 1))
    rlim = kwargs.get('rlim', (0, 15))
    clim = kwargs.get('clim', (0, 1))

    fig = pl.figure(figsize=(fw, fw / 4))
    gs_tt = GridSpec(2, 1, bottom=0.2, top=1, left=0.1, right=0.3, hspace=0, wspace=0, height_ratios=[0.15, 0.85])
    gs_ct = GridSpec(2, 5, bottom=0.2, top=1, left=0.37, right=1, hspace=0.05, wspace=0.05,
                     height_ratios=[0.15, 0.85],
                     width_ratios=[1, 1, 1, 1, 0.2])

    ax_tt = pl.subplot(gs_tt[1, 0])
    ax_chj = pl.subplot(gs_ct[1, 0])
    ax_ccj = pl.subplot(gs_ct[1, 1])
    ax_cbj = pl.subplot(gs_ct[1, 2])
    ax_crj = pl.subplot(gs_ct[1, 3])
    ax_thm = pl.subplot(gs_ct[0, 0])
    ax_ctm = pl.subplot(gs_ct[0, 1])
    ax_bm = pl.subplot(gs_ct[0, 2])
    ax_rm = pl.subplot(gs_ct[0, 3])
    ax_cnm = pl.subplot(gs_ct[1, 4])

    ax_tt.hexbin(hte, cte, gridsize=gs, cmap=cmap, extent=(htelim[0], htelim[1], ctelim[0], ctelim[1]))
    ax_chj.hexbin(hte, cnr, gridsize=gs, cmap=cmap, extent=(htelim[0], htelim[1], clim[0], clim[1]))
    ax_ccj.hexbin(cte, cnr, gridsize=gs, cmap=cmap, extent=(ctelim[0], ctelim[1], clim[0], clim[1]))
    ax_cbj.hexbin(imp, cnr, gridsize=gs, cmap=cmap, extent=(blim[0], blim[1], clim[0], clim[1]))
    ax_crj.hexbin(rho, cnr, gridsize=gs, cmap=cmap, extent=(rlim[0], rlim[1], clim[0], clim[1]))

    ax_thm.hist(hte, bins=nb, alpha=0.5, range=htelim)
    ax_ctm.hist(cte, bins=nb, alpha=0.5, range=ctelim)
    ax_bm.hist(imp, bins=nb, alpha=0.5, range=blim)
    ax_rm.hist(rho, bins=nb, alpha=0.5, range=rlim)
    ax_cnm.hist(cnr, bins=nb, alpha=0.5, range=clim, orientation='horizontal')

    pl.setp(ax_tt, xlabel='Host $T_\mathrm{Eff}$', ylabel='Contaminant $T_\mathrm{Eff}$')
    pl.setp(ax_chj, xlabel='Host $T_\mathrm{Eff}$', ylabel='Contamination in $i\'$')
    pl.setp(ax_ccj, xlabel='Contaminant $T_\mathrm{Eff}$')
    pl.setp(ax_cbj, xlabel='Impact parameter')
    pl.setp(ax_crj, xlabel='Stellar density')

    pl.setp(ax_thm, xlim=ax_chj.get_xlim())
    pl.setp(ax_ctm, xlim=ax_ccj.get_xlim())
    pl.setp(ax_bm, xlim=ax_cbj.get_xlim())
    pl.setp([ax_ccj, ax_cnm], ylim=ax_chj.get_ylim())
    pl.setp([ax_chj, ax_ccj, ax_cbj, ax_crj, ax_cnm], ylim=clim)

    pl.setp([ax_thm, ax_ctm, ax_cnm, ax_bm, ax_rm], yticks=[], xticks=[])
    pl.setp(ax_ccj.get_yticklabels(), visible=False)
    pl.setp(ax_cbj.get_yticklabels(), visible=False)
    pl.setp(ax_crj.get_yticklabels(), visible=False)
    [sb.despine(ax=ax, left=True, offset=0.1) for ax in [ax_thm, ax_ctm, ax_bm, ax_rm]]
    [sb.despine(ax=ax) for ax in [ax_chj, ax_ccj, ax_cbj, ax_crj]]
    sb.despine(ax=ax_cnm, bottom=True)
    return fig, ax_tt, ax_chj, ax_cbj, ax_ccj, ax_crj

def joint_marginal_plot(df, fw=10, nb=30, gs=25, **kwargs):
    return _jplot(df.teff_h, df.teff_c, df.cnt, df.b, df.rho, fw, nb, gs, **kwargs)[0]



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
from typing import List, Union, Iterable, Sequence, Optional

from astropy.stats import sigma_clip
from matplotlib.pyplot import subplots, setp
from numba import njit, prange
from numpy import (inf, sqrt, ones, zeros_like, concatenate, diff, log, ones_like, all,
                   clip, argsort, any, s_, zeros, arccos, nan, full, pi, sum, repeat, asarray, ndarray, log10,
                   array, atleast_2d, isscalar, atleast_1d, where, isfinite, arange, unique, squeeze, ceil, percentile,
                   floor, diag, nanstd)
from numpy.random import uniform, normal, permutation, multivariate_normal
from scipy.stats import norm

from .loglikelihood import WNLogLikelihood, CeleriteLogLikelihood

try:
    from ldtk import LDPSetCreator
    with_ldtk = True
except ImportError:
    with_ldtk = False

from ..models.transitmodel import TransitModel
from ..orbits.orbits_py import duration_eccentric, as_from_rhop, i_from_ba, i_from_baew, d_from_pkaiews, epoch
from ..param.parameter import ParameterSet, PParameter, GParameter, LParameter
from ..param.parameter import UniformPrior as U, NormalPrior as N, GammaPrior as GM
from .. import QuadraticModel
from .logposteriorfunction import LogPosteriorFunction


@njit(cache=False)
def lnlike_normal(o, m, e):
    return -sum(log(e)) -0.5*o.size*log(2.*pi) - 0.5*sum((o-m)**2/e**2)


@njit(cache=False)
def lnlike_normal_s(o, m, e):
    return -o.size*log(e) -0.5*o.size*log(2.*pi) - 0.5*sum((o-m)**2)/e**2


@njit(parallel=True, cache=False, fastmath=True)
def lnlike_normal_v(o, m, e, wnids, lcids):
    m = atleast_2d(m)
    npv = m.shape[0]
    npt = o.size
    lnl = zeros(npv)
    for i in prange(npv):
        for j in range(npt):
            k = wnids[lcids[j]]
            lnl[i] += -log(e[i,k]) - 0.5*log(2*pi) - 0.5*((o[j]-m[i,j])/e[i,k])**2
    return lnl


@njit(fastmath=True)
def map_pv(pv):
    pv = atleast_2d(pv)
    pvt = zeros((pv.shape[0], 7))
    pvt[:,0]   = sqrt(pv[:,4])
    pvt[:,1:3] = pv[:,0:2]
    pvt[:,  3] = as_from_rhop(pv[:,2], pv[:,1])
    pvt[:,  4] = i_from_ba(pv[:,3], pvt[:,3])
    return pvt


@njit(fastmath=True, cache=False)
def map_ldc(ldc):
    ldc = atleast_2d(ldc)
    uv = zeros_like(ldc)
    a, b = sqrt(ldc[:,0::2]), 2.*ldc[:,1::2]
    uv[:,0::2] = a * b
    uv[:,1::2] = a * (1. - b)
    return uv


class BaseLPF(LogPosteriorFunction):
    _lpf_name = 'BaseLPF'

    def __init__(self, name: str,
                 passbands: Union[Sequence[str], str],
                 times: Optional[Sequence[ndarray]] = None,
                 fluxes: Optional[Sequence[ndarray]] = None,
                 errors: Optional[Sequence[ndarray]]  = None,
                 pbids: Optional[Sequence[int]]  = None,
                 covariates: Optional[Sequence[ndarray]]  = None,
                 wnids: Optional[Sequence[int]]  = None,
                 tm: Optional[TransitModel] = None,
                 nsamples: Union[Sequence[int], int] = 1,
                 exptimes: Union[Sequence[float], float] = 0.0,
                 init_data: bool = True, result_dir: Optional[Path] = None, tref: float = 0.0,
                 lnlikelihood: str = 'wn'):
        """The base Log Posterior Function class.

        The `BaseLPF` class creates the basis for transit light curve analyses using `PyTransit`. This class can be
        used in a basic analysis directly, or it can be inherited to create a basis for a more complex analysis.

        Parameters
        ----------
        name
            Name of the log posterior function instance.

        passbands
            A passband name or a sequence of unique passband names (filters) that the light curves have been observed in.

        times
            A sequence of 1d ndarrays each containing the mid-observation times for a single light curve.

        fluxes
            A sequence of 1d ndarrays each containing the normalized fluxes for a single light curve.

        errors
            A sequence of 1d ndarrays each containing the flux measurement uncertainties for a single light curvel.

        pbids
            A sequence of passband indices mapping each light curve to a single passband.

        covariates
            A sequence of covariates, one 2d ndarray per light curve.

        wnids
            A sequence of noise set indices mapping each light curve to a single noise set.

        tm
            The transit model to use instead of the default QuadraticModel.

        nsamples: Sequence[int]
            A sequence of supersampling factors. The values should be integers and given per light curve.

        exptimes
            A sequence of exposure times. The values should be floats with the time given in days.

        init_data
            Set to `False` to allow the LPF to be initialized without data. This is mainly for debugging.

        result_dir
            Default saving directory

        tref
            Reference time.

        lnlikelihood
            The log-likelihood model to use. Can be either 'wn' or 'celerite'.
        """

        self._pre_initialisation()

        super().__init__(name=name, result_dir=result_dir)

        self.tm: TransitModel = tm or QuadraticModel(klims=(0.01, 0.75), nk=512, nz=512)
        self._tref: float = tref

        # Passbands
        # ---------
        # Passbands should be arranged from blue to red
        if isinstance(passbands, (list, tuple, ndarray)):
            self.passbands: list = list(passbands)
        else:
            self.passbands: list = [passbands]
        self.npb = npb = len(self.passbands)

        self.nsamples = None
        self.exptimes = None
        self.lnlikelihood_type = lnlikelihood.lower()

        # Declare high-level objects
        # --------------------------
        self._lnlikelihood_models: list = []
        self._baseline_models: list = []
        self.ps = None          # Parametrisation
        self.de = None          # Differential evolution optimiser
        self.sampler = None     # MCMC sampler
        self.instrument = None  # Instrument
        self.ldsc = None        # Limb darkening set creator
        self.ldps = None        # Limb darkening profile set
        self.cntm = None        # Contamination model

        # Declare data arrays and variables
        # ---------------------------------
        self.nlc: int = 0                # Number of light curves
        self.n_noise_blocks: int = 0     # Number of noise blocks
        self.noise_ids = None
        self.times: Optional[list] = None          # List of time arrays
        self.fluxes: Optional[list] = None         # List of flux arrays
        self.errors: Optional[list] = None         # List of flux uncertainties
        self.covariates: Optional[list] = None     # List of covariates
        self.wn: Optional[ndarray] = None          # Array of white noise estimates for each light curve
        self.timea: Optional[ndarray] = None       # Array of concatenated times
        self.mfluxa: Optional[ndarray] = None      # Array of concatenated model fluxes
        self.ofluxa: Optional[ndarray] = None      # Array of concatenated observed fluxes
        self.errora: Optional[ndarray] = None      # Array of concatenated model fluxes

        self.lcids: Optional[ndarray] = None       # Array of light curve indices for each datapoint
        self.pbids: Optional[ndarray] = None       # Array of passband indices for each light curve
        self.lcslices: Optional[list] = None       # List of light curve slices

        if init_data:
            # Set up the observation data
            # ---------------------------
            self._init_data(times = times, fluxes = fluxes, pbids = pbids, covariates = covariates,
                            errors = errors, wnids = wnids, nsamples = nsamples, exptimes = exptimes)

            # Set up the parametrisation
            # --------------------------
            self._init_parameters()

            # Inititalise the instrument
            # --------------------------
            self._init_instrument()

        self._init_lnlikelihood()
        self._init_baseline()
        self._post_initialisation()


    def _init_data(self, times: Union[Sequence, ndarray], fluxes: Union[Sequence, ndarray], pbids: Union[Sequence, ndarray] = None,
                   covariates: Union[Sequence, ndarray] = None, errors: Union[Sequence, ndarray] = None, wnids: Union[Sequence, ndarray] = None,
                   nsamples: Union[int, ndarray, Sequence] = 1, exptimes: Union[float, ndarray, Sequence] = 0.):

        if isinstance(times, ndarray) and times.ndim == 1 and times.dtype == float:
            times = [times]

        if isinstance(fluxes, ndarray) and fluxes.ndim == 1 and fluxes.dtype == float:
            fluxes = [fluxes]

        if pbids is None:
            if self.pbids is None:
                self.pbids = zeros(len(fluxes), int)
        else:
            self.pbids = atleast_1d(pbids).astype('int')

        self.nlc = len(times)
        self.times = times
        self.fluxes = fluxes
        self.wn = [nanstd(diff(f)) / sqrt(2) for f in fluxes]
        self.timea = concatenate(self.times)
        self.ofluxa = concatenate(self.fluxes)
        self.mfluxa = zeros_like(self.ofluxa)
        self.lcids = concatenate([full(t.size, i) for i, t in enumerate(self.times)])

        # TODO: Noise IDs get scrambled when removing transits, fix!!!
        if wnids is None:
            if self.noise_ids is None:
                self.noise_ids = zeros(self.nlc, int)
                self.n_noise_blocks = 1
        else:
            self.noise_ids = asarray(wnids)
            self.n_noise_blocks = len(unique(self.noise_ids))
            assert self.noise_ids.size == self.nlc, "Need one noise block id per light curve."
            assert self.noise_ids.max() == self.n_noise_blocks - 1, "Error initialising noise block ids."

        if isscalar(nsamples):
            self.nsamples = full(self.nlc, nsamples)
            self.exptimes = full(self.nlc, exptimes)
        else:
            assert (len(nsamples) == self.nlc) and (len(exptimes) == self.nlc)
            self.nsamples = asarray(nsamples, 'int')
            self.exptimes = asarray(exptimes)

        self.tm.set_data(self.timea-self._tref, self.lcids, self.pbids, self.nsamples, self.exptimes)

        if errors is None:
            self.errors = [full(t.size, nan) for t in self.times]
        else:
            self.errors = errors
        self.errora = concatenate(self.errors)

        # Initialise the light curves slices
        # ----------------------------------
        self.lcslices = []
        sstart = 0
        for i in range(self.nlc):
            s = self.times[i].size
            self.lcslices.append(s_[sstart:sstart + s])
            sstart += s

        # Initialise the covariate arrays, if given
        # -----------------------------------------
        if covariates is not None:
            self.covariates = covariates
            for cv in self.covariates:
                cv = (cv - cv.mean(0)) / cv.std(0)
            #self.ncovs = self.covariates[0].shape[1]
            #self.covsize = array([c.size for c in self.covariates])
            #self.covstart = concatenate([[0], self.covsize.cumsum()[:-1]])
            #self.cova = concatenate(self.covariates)

    def _add_lnlikelihood_model(self, lnl):
        self._lnlikelihood_models.append(lnl)

    def _add_baseline_model(self, blm):
        self._baseline_models.append(blm)

    def _init_parameters(self):
        self.ps = ParameterSet()
        self._init_p_orbit()
        self._init_p_planet()
        self._init_p_limb_darkening()
        self._init_p_baseline()
        self.ps.freeze()

    def _init_p_orbit(self):
        """Orbit parameter initialisation.
        """
        porbit = [
            GParameter('tc',  'zero_epoch',       'd',      N(0.0,  0.1), (-inf, inf)),
            GParameter('p',   'period',           'd',      N(1.0, 1e-5), (0,    inf)),
            GParameter('rho', 'stellar_density',  'g/cm^3', U(0.1, 25.0), (0,    inf)),
            GParameter('b',   'impact_parameter', 'R_s',    U(0.0,  1.0), (0,      1))]
        self.ps.add_global_block('orbit', porbit)

    def _init_p_planet(self):
        """Planet parameter initialisation.
        """
        pk2 = [PParameter('k2', 'area_ratio', 'A_s', U(0.05**2, 0.2**2), (0, inf))]
        self.ps.add_passband_block('k2', 1, 1, pk2)
        self._pid_k2 = repeat(self.ps.blocks[-1].start, self.npb)
        self._start_k2 = self.ps.blocks[-1].start
        self._sl_k2 = self.ps.blocks[-1].slice

    def _init_p_limb_darkening(self):
        """Limb darkening parameter initialisation.
        """
        pld = concatenate([
            [PParameter(f'q1_{pb}', 'q1 coefficient {pb}', '', U(0, 1), bounds=(0, 1)),
             PParameter(f'q2_{pb}', 'q2 coefficient {pb}', '', U(0, 1), bounds=(0, 1))]
            for i,pb in enumerate(self.passbands)])
        self.ps.add_passband_block('ldc', 2, self.npb, pld)
        self._sl_ld = self.ps.blocks[-1].slice
        self._start_ld = self.ps.blocks[-1].start

    def _init_p_baseline(self):
        pass

    def _init_p_noise(self):
        pass

    def _init_instrument(self):
        pass

    def _pre_initialisation(self):
        pass

    def _post_initialisation(self):
        pass

    def _init_lnlikelihood(self):
        if self.lnlikelihood_type == 'wn':
            self._add_lnlikelihood_model(WNLogLikelihood(self))
        elif self.lnlikelihood_type == 'celerite':
            self._add_lnlikelihood_model(CeleriteLogLikelihood(self))
        else:
            raise NotImplementedError

    def _init_baseline(self):
        pass

    def create_pv_population(self, npop=50):
        pvp = self.ps.sample_from_prior(npop)

        # With LDTk
        # ---------
        #
        # Use LDTk to create the sample if LDTk has been initialised.
        #
        if self.ldps:
            istart = self._start_ld
            cms, ces = self.ldps.coeffs_tq()
            for i, (cm, ce) in enumerate(zip(cms.flat, ces.flat)):
                pvp[:, i + istart] = normal(cm, ce, size=pvp.shape[0])

        # No LDTk
        # -------
        #
        # Ensure that the total limb darkening decreases towards
        # red passbands.
        #
        else:
            ldsl = self._sl_ld
            for i in range(pvp.shape[0]):
                pid = argsort(pvp[i, ldsl][::2])[::-1]
                pvp[i, ldsl][::2] = pvp[i, ldsl][::2][pid]
                pvp[i, ldsl][1::2] = pvp[i, ldsl][1::2][pid]

        return pvp

    def add_prior(self, prior):
        self._additional_log_priors.append(prior)

    def baseline(self, pv):
        if self._baseline_models:
            if pv.ndim == 1:
                bl = ones_like(self.timea)
            else:
                bl = ones((pv.shape[0], self.timea.size))
            for blm in self._baseline_models:
                bl = blm(pv, bl)
            return bl
        else:
            return 1.

    def trends(self, pv):
        """Additive trends"""
        return 0.

    def transit_model(self, pv, copy=True):
        pv = atleast_2d(pv)
        ldc = map_ldc(pv[:,self._sl_ld])
        zero_epoch = pv[:,0] - self._tref
        period = pv[:,1]
        smaxis = as_from_rhop(pv[:,2], period)
        inclination  = i_from_ba(pv[:,3], smaxis)
        radius_ratio = sqrt(pv[:,4:5])
        return self.tm.evaluate(radius_ratio, ldc, zero_epoch, period, smaxis, inclination)

    def flux_model(self, pv):
        baseline    = self.baseline(pv)
        trends      = self.trends(pv)
        model_flux = self.transit_model(pv)
        return baseline * model_flux + trends

    def residuals(self, pv):
        return self.ofluxa - self.flux_model(pv)

    def lnlikelihood(self, pvp):
        """Log likelihood for a 1D or 2D array of model parameters.

        Parameters
        ----------
        pvp: ndarray
            Either a 1D parameter vector or a 2D parameter array.

        Returns
        -------
            Log likelihood for the given parameter vector(s).
        """
        fmodel = self.flux_model(pvp)

        if pvp.ndim == 1:
            lnl = 0.
        else:
            lnl = zeros(pvp.shape[0])

        for lnlikelihood in self._lnlikelihood_models:
            lnl += lnlikelihood(pvp, fmodel)
        return lnl

    def set_radius_ratio_prior(self, kmin, kmax):
        """Set a uniform prior on all radius ratios."""
        for p in self.ps[self._sl_k2]:
            p.prior = U(kmin ** 2, kmax ** 2)

    def add_t14_prior(self, mean: float, std: float) -> None:
        """Add a normal prior on the transit duration.

        Parameters
        ----------
        mean: float
            Mean of the normal distribution
        std: float
            Standard deviation of the normal distribution.
        """

        def T14(pv):
            pv = atleast_2d(pv)
            a = as_from_rhop(pv[:, 2], pv[:, 1])
            t14 = duration_eccentric(pv[:, 1], sqrt(pv[:, 4]), a, arccos(pv[:, 3] / a), 0, 0, 1)
            return norm.logpdf(t14, mean, std)

        self._additional_log_priors.append(T14)

    def add_as_prior(self, mean: float, std: float) -> None:
        """Add a normal prior on the scaled semi-major axis :math:`(a / R_\star)`.

        Parameters
        ----------
        mean: float
            Mean of the normal distribution.
        std: float
            Standard deviation of the normal distribution
        """
        def as_prior(pv):
            a = as_from_rhop(pv[2], pv[1])
            return norm.logpdf(a, mean, std)
        self._additional_log_priors.append(as_prior)

    def add_ldtk_prior(self, teff: tuple, logg: tuple, z: tuple, passbands: tuple,
                       uncertainty_multiplier: float = 3, **kwargs) -> None:
        """Add a LDTk-based prior on the limb darkening.

        Parameters
        ----------
        teff
        logg
        z
        passbands
        uncertainty_multiplier

        Returns
        -------

        """
        if 'pbs' in kwargs.keys():
            raise DeprecationWarning("The 'pbs' argument has been renamed to 'passbands'")

        if isinstance(passbands[0], str):
            raise DeprecationWarning(
                'Passing passbands by name has been deprecated, they should be now Filter instances.')

        self.ldsc = LDPSetCreator(teff, logg, z, list(passbands))
        self.ldps = self.ldsc.create_profiles(1000)
        self.ldps.resample_linear_z()
        self.ldps.set_uncertainty_multiplier(uncertainty_multiplier)
        def ldprior(pv):
            return self.ldps.lnlike_tq(pv[:, self._sl_ld].reshape([pv.shape[0], -1, 2]))
        self._additional_log_priors.append(ldprior)

    def remove_outliers(self, sigma=5):
        fmodel = squeeze(self.flux_model(self.de.minimum_location))
        covariates = [] if self.covariates is not None else None
        times, fluxes, lcids, errors = [], [], [], []
        for i in range(len(self.times)):
            res = self.fluxes[i] - fmodel[self.lcslices[i]]
            mask = ~sigma_clip(res, sigma=sigma).mask
            times.append(self.times[i][mask])
            fluxes.append(self.fluxes[i][mask])
            if covariates is not None:
                covariates.append(self.covariates[i][mask])
            if self.errors is not None:
                errors.append(self.errors[i][mask])

        self._init_data(times=times, fluxes=fluxes, covariates=self.covariates, pbids=self.pbids,
                        errors=(errors if self.errors is not None else None), wnids=self.noise_ids,
                        nsamples=self.nsamples, exptimes=self.exptimes)

    def remove_transits(self, tids):
        m = ones(len(self.times), bool)
        m[tids] = False
        self._init_data(self.times[m], self.fluxes[m], self.pbids[m],
                        self.covariates[m] if self.covariates is not None else None,
                        self.errors[m], self.noise_ids[m], self.nsamples[m], self.exptimes[m])
        self._init_parameters()

    def posterior_samples(self, burn: int = 0, thin: int = 1, derived_parameters: bool = True):
        df = super().posterior_samples(burn=burn, thin=thin)
        if derived_parameters:
            for k2c in df.columns[self._sl_k2]:
                df[k2c.replace('k2', 'k')] = sqrt(df[k2c])
            df['a'] = as_from_rhop(df.rho.values, df.p.values)
            df['inc'] = i_from_baew(df.b.values, df.a.values, 0., 0.)

            average_ks = sqrt(df.iloc[:, self._sl_k2]).mean(1).values
            df['t14'] = d_from_pkaiews(df.p.values, average_ks, df.a.values, df.inc.values, 0., 0., 1, kind=14)
            df['t23'] = d_from_pkaiews(df.p.values, average_ks, df.a.values, df.inc.values, 0., 0., 1, kind=23)
        return df

    def plot_light_curves(self, method='de', ncol: int = 3, width: float = 2., max_samples: int = 1000, figsize=None,
                          data_alpha=0.5, ylim=None):
        nrow = int(ceil(self.nlc / ncol))
        if method == 'mcmc':
            df = self.posterior_samples(derived_parameters=False)
            t0, p = df.tc.median(), df.p.median()
            fmodel = self.flux_model(permutation(df.values)[:max_samples])
            fmperc = percentile(fmodel, [50, 16, 84, 2.5, 97.5], 0)
        else:
            fmodel = squeeze(self.flux_model(self.de.minimum_location))
            t0, p = self.de.minimum_location[0], self.de.minimum_location[1]
            fmperc = None

        fig, axs = subplots(nrow, ncol, figsize=figsize, constrained_layout=True, sharey='all', sharex='all',
                            squeeze=False)
        for i in range(self.nlc):
            ax = axs.flat[i]
            e = epoch(self.times[i].mean(), t0, p)
            tc = t0 + e * p
            time = self.times[i] - tc

            ax.plot(time, self.fluxes[i], '.', alpha=data_alpha)

            if method == 'de':
                ax.plot(time, fmodel[self.lcslices[i]], 'w', lw=4)
                ax.plot(time, fmodel[self.lcslices[i]], 'k', lw=1)
            else:
                ax.fill_between(time, *fmperc[3:5, self.lcslices[i]], alpha=0.15)
                ax.fill_between(time, *fmperc[1:3, self.lcslices[i]], alpha=0.25)
                ax.plot(time, fmperc[0, self.lcslices[i]])

            setp(ax, xlabel=f'Time - T$_c$ [d]', xlim=(-width / 2 / 24, width / 2 / 24))
        setp(axs[:, 0], ylabel='Normalised flux')

        if ylim is not None:
            setp(axs, ylim=ylim)

        for ax in axs.flat[self.nlc:]:
            ax.remove()
        return fig

    def __repr__(self):
        return f"Target: {self.name}\nLPF: {self._lpf_name}\n Passbands: {self.passbands}"



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import njit, prange
from scipy.constants import G, k, h, c
from numpy import exp, pi, sqrt, zeros, sin, cos, nan, inf, linspace, meshgrid, floor, isfinite, fmax, isnan, nanmean, \
    arange, zeros_like, atleast_2d, polyfit
from scipy.interpolate import interp1d

from .rrmodel import circle_circle_intersection_area
from ...contamination.contamination import read_phoenix_spectrum_table
from ...orbits.taylor_z import vajs_from_paiew, find_contact_point

d2sec = 24.*60.*60.


@njit
def polyflux_multiband(t, tref, coeffs):
    coeffs = atleast_2d(coeffs)
    npb = coeffs.shape[0]
    fluxes = zeros(npb)

    tt = t - tref
    for i in range(npb):
        c = coeffs[i]
        fluxes[i] = c[5] + c[4] * tt + c[3] * tt ** 2 + c[2] * tt * 3 + c[1] * tt ** 4 + c[0] * tt ** 5
    return fluxes


@njit
def polyflux_singleband(t, tref, ipb, coeffs):
    c = atleast_2d(coeffs)[ipb]
    tt = t - tref
    return c[5] + c[4] * tt + c[3] * tt ** 2 + c[2] * tt * 3 + c[1] * tt ** 4 + c[0] * tt ** 5


@njit
def planck(l, t):
    return 2.*h*c**2/l**5/(exp(h*c/(l*k*t)) - 1.)


@njit
def stellar_oblateness(w, rho):
    return 3.*w*w/(8.*G*pi*rho)


@njit
def z_s(x, y, r, f, sphi, cphi):
    x2, y2, r2 = x*x, y*y, r*r
    if x2 + y2 > r2:
        return -1.0
    else:
        sphi2, cphi2, f2 = sphi**2, cphi**2, f**2

        a = 1./(1. - f)**2
        d = (4.*y2*sphi2*cphi2*(a - 1.)**2- 4.*(cphi2 + a*sphi2)*(x2 + y2*(a*cphi2 + sphi2) - r2))
        if d >= 0.0:
            return (-2.*y*cphi*sphi*(a - 1.) + sqrt(d))/(2.*(cphi2 + a*sphi2))
        else:
            return -1.0


@njit
def z_v(xs, ys, r, f, sphi, cphi):
    npt = xs.size
    z = zeros(npt)

    sphi2, cphi2, f2, r2 = sphi**2, cphi**2, f**2, r**2
    a = 1./(1. - f)**2

    for i in range(npt):
        x2, y2 = xs[i]**2, ys[i]**2
        if x2 + y2 > r2:
            z[i] = -1.0
        else:
            d = (4.*y2*sphi2*cphi2*(a - 1.)**2 - 4.*(cphi2 + a*sphi2)*(x2 + y2*(a*cphi2 + sphi2) - r2))
            if d >= 0.0:
                z[i] = (-2.*ys[i]*cphi*sphi*(a - 1.) + sqrt(d))/(2.*(cphi2 + a*sphi2))
            else:
                z[i] = -1.0
    return z


@njit
def luminosity_s(x, y, mstar, rstar, ostar, tpole, gpole, f, sphi, cphi, beta, ldc, tref, ipb, fcoeffs):
    dg = zeros(3)
    dc = zeros(3)

    z = z_s(x, y, rstar, f, sphi, cphi)
    if z < 0.0:
        return nan
    else:
        mu = z/sqrt((x**2 + y**2 + z**2))

        dg[0] = x
        dg[1] = y*cphi + z*sphi
        dg[2] = -y*sphi + z*cphi

        dc[0] = dg[0]
        dc[2] = dg[2]

        # Direction vector lengths
        # ------------------------
        lg2 = dg[0]**2 + dg[1]**2 + dg[2]**2
        lg = sqrt(lg2)
        lc = sqrt(dc[0]**2 + dc[2]**2)

        # Normalize the direction vectors
        # -------------------------------
        dg /= lg
        dc /= lc

        gg = - G*mstar/lg2
        gc = ostar*ostar*lc

        dgg = gg*dg + gc*dc
        g = sqrt((dgg**2).sum())
        t = tpole*g**beta/gpole**beta

        return polyflux_singleband(t, tref, ipb, fcoeffs) * (1. - ldc[0] * (1. - mu) - ldc[1] * (1. - mu) ** 2)


@njit
def luminosity_v(xs, ys, mstar, rstar, ostar, tpole, gpole, f, sphi, cphi, beta, ldc, tref, ipb, fcoeffs):
    npt = xs.size
    l = zeros(npt)
    dg = zeros(3)
    dc = zeros(3)

    for i in range(npt):
        x, y = xs[i], ys[i]
        z = z_s(x, y, rstar, f, sphi, cphi)
        if z < 0.0:
            l[i] = nan
        else:
            mu = z/sqrt((x**2 + y**2 + z**2))

            dg[0] = x
            dg[1] = y*cphi + z*sphi
            dg[2] = -y*sphi + z*cphi

            dc[0] = dg[0]
            dc[2] = dg[2]

            # Direction vector lengths
            # ------------------------
            lg2 = dg[0]**2 + dg[1]**2 + dg[2]**2
            lg = sqrt(lg2)
            lc = sqrt(dc[0]**2 + dc[2]**2)

            # Normalize the direction vectors
            # -------------------------------
            dg /= lg
            dc /= lc

            gg = - G*mstar/lg2
            gc = ostar*ostar*lc

            dgg = gg*dg + gc*dc
            g = sqrt((dgg**2).sum())
            t = tpole*g**beta/gpole**beta

            l[i] = polyflux_singleband(t, tref, ipb, fcoeffs) * (1. - ldc[0]*(1. - mu) - ldc[1]*(1. - mu)**2)
    return l


@njit
def luminosity_v2(ps, normals, istar, mstar, rstar, ostar, tpole, gpole, beta, ldc, tref, ipb, fcoeffs):
    npt = ps.shape[0]
    l = zeros(npt)
    dc = zeros(3)

    vx = 0.0
    vy = -cos(istar)
    vz = -sin(istar)

    for i in range(npt):
        px, py, pz = ps[i] * rstar         # Position vector components
        nx, ny, nz = normals[i]            # Normal vector components

        mu =  vy*ny + vz*nz

        lp2 = (px**2 + py**2 + pz**2)      # Squared distance from center
        lc = sqrt(px**2 + pz**2)           # Centrifugal vector length
        cx, cz = px/lc, pz/lc              # Normalized centrifugal vector

        gg = -G * mstar / lp2              # Newtonian surface gravity component
        gc = ostar * ostar * lc            # Centrifugal surface gravity component

        gx = gg*nx + gc*cx                 # Surface gravity x component
        gy = gg*ny                         # Surface gravity y component
        gz = gg*nz + gc*cz                 # Surface gravity z component

        g = sqrt((gx**2 + gy**2 + gz**2))  # Surface gravity
        t = tpole*g**beta / gpole**beta    # Temperature [K]
        l[i] = polyflux_singleband(t, tref, ipb, fcoeffs) # Thermal radiation
        l[i] *= (1.-ldc[0]*(1.-mu) - ldc[1]*(1.-mu)**2) # Quadratic limb darkening
    return l


@njit
def luminosity_s2(p, normal, istar, mstar, rstar, ostar, tpole, gpole, beta, ldc,  tref, ipb, fcoeffs):

    vx = 0.0
    vy = -cos(istar)
    vz = -sin(istar)

    px, py, pz = p * rstar             # Position vector components
    nx, ny, nz = normal                # Normal vector components

    mu =  vy*ny + vz*nz

    lp2 = (px**2 + py**2 + pz**2)      # Squared distance from center
    lc = sqrt(px**2 + pz**2)           # Centrifugal vector length
    cx, cz = px/lc, pz/lc              # Normalized centrifugal vector

    gg = -G * mstar / lp2              # Newtionian surface gravity component
    gc = ostar * ostar * lc            # Centrifugal surface gravity component

    gx = gg*nx + gc*cx                 # Surface gravity x component
    gy = gg*ny                         # Surface gravity y component
    gz = gg*nz + gc*cz                 # Surface gravity z component

    g = sqrt((gx**2 + gy**2 + gz**2))  # Surface gravity
    t = tpole*g**beta / gpole**beta    # Temperature [K]
    l = polyflux_singleband(t, tref, ipb, fcoeffs)  # Thermal radiation
    l *= (1.-ldc[0]*(1.-mu) - ldc[1]*(1.-mu)**2) # Quadratic limb darkening
    return l


def create_star_xy(res: int = 64):
    st = linspace(-1., 1., res)
    x, y = meshgrid(st, st)
    return st, x.ravel(), y.ravel()


def create_planet_xy(res: int = 6):
    dd = 2/(res + 1)
    dt = dd*arange(1, res + 1) - 1
    xs, ys = meshgrid(dt, dt)
    m = sqrt(xs**2 + ys**2) <= 1.
    return xs[m], ys[m]


@njit
def create_star_luminosity(res, x, y, mstar, rstar, ostar, tpole, gpole, f, sphi, cphi, beta, ldc, tref, fcoeffs):
    l = zeros((fcoeffs.shape[0], res, res))
    for ipb in range(fcoeffs.shape[0]):
        l[ipb] = luminosity_v(x*rstar, y*rstar, mstar, rstar, ostar, tpole, gpole, f, sphi, cphi, beta, ldc, tref, ipb, fcoeffs).reshape((res,res))
    return l


@njit(cache=False, fastmath=False)
def mean_luminosity(xc, yc, k, xs, ys, feff, lt, xt, yt):
    ns = xs.size
    rf = 1.0/(1.0 - feff)

    lsum = 0.0
    weight = 0.0
    for i in range(ns):
        x = xc + xs[i]*k
        y = (yc + ys[i]*k)*rf

        if x**2 + y**2 > 1.0:
            continue

        dx = xt[1] - xt[0]
        dy = yt[1] - yt[0]

        ix = int(floor((x - xt[0])/dx))
        ax1 = (x - xt[ix])/dx
        ax2 = 1.0 - ax1

        iy = int(floor((y - yt[0])/dy))
        ay1 = (y - yt[iy])/dy
        ay2 = 1.0 - ay1

        l = (  lt[iy,     ix    ]*ay2*ax2
             + lt[iy + 1, ix    ]*ay1*ax2
             + lt[iy,     ix + 1]*ay2*ax1
             + lt[iy + 1, ix + 1]*ay1*ax1)

        if isfinite(l):
            lsum += l
            weight += 1.

    if weight > 0.:
        return lsum/weight
    else:
        return nan


@njit
def mean_luminosity_under_planet(x, y, mstar, rstar, ostar, tpole, gpole, f, sphi, cphi, beta, ldc, tref, ipb, fcoeffs):
    l = luminosity_v(x*rstar, y*rstar, mstar, rstar, ostar, tpole, gpole, f, sphi, cphi, beta, ldc, tref, ipb, fcoeffs)
    return nanmean(l)


@njit
def calculate_luminosity_interpolation_table(res, k, xp, yp, sa, ca, y0, vx, vy, ax, ay, jx, jy, sx, sy,
                                             mstar, rstar, ostar, tpole, gpole, f, feff, sphi, cphi, beta, ldc,
                                             tref, fcoeffs):
    t1 = find_contact_point(k, 1, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    t4 = find_contact_point(k, 4, y0, vx, vy, ax, ay, jx, jy, sx, sy)
    times = linspace(t1, t4, res)

    npb = fcoeffs.shape[0]
    lt = zeros((npb, res))
    for ipb in range(npb):
        for i in range(lt.size):
            x, y = xy_taylor_st(times[i], sa, ca, y0, vx, vy, ax, ay, jx, jy, sx, sy)
            xs = x + k * xp
            ys = y + k * yp
            lt[ipb, i] = mean_luminosity_under_planet(xs, ys, mstar, rstar, ostar, tpole, gpole, f, sphi, cphi, beta, ldc,
                                                 tref, ipb, fcoeffs)

        i = 0
        while isnan(lt[ipb, i]):
            i += 1

        ifill = lt[ipb, i]
        while i >= 0:
            lt[ipb, i] = ifill
            i -= 1

        i = lt.shape[1] - 1
        while isnan(lt[ipb, i]):
            i -= 1

        ifill = lt[ipb, i]
        while i < lt.shape[1]:
            lt[ipb, i] = ifill
            i += 1

    return times, lt


@njit(fastmath=True)
def xy_taylor_st(t, sa, ca, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    t2 = t*t
    t3 = t2*t
    t4 = t3*t
    px =      vx*t + 0.5*ax*t2 + jx*t3/6.0 + sx*t4/24.
    py = y0 + vy*t + 0.5*ay*t2 + jy*t3/6.0 + sy*t4/24.

    x = ca*px - sa*py
    y = ca*py + sa*px

    return x, y


@njit(fastmath=True)
def xy_taylor_vt(ts, a, y0, vx, vy, ax, ay, jx, jy, sx, sy):
    npt = ts.size
    x, y = zeros(npt), zeros(npt)
    ca, sa = cos(a), sin(a)

    for i in range(npt):
        t = ts[i]
        t2 = t*t
        t3 = t2*t
        t4 = t3*t
        px =      vx*t + 0.5*ax*t2 + jx*t3/6.0 + sx*t4/24.
        py = y0 + vy*t + 0.5*ay*t2 + jy*t3/6.0 + sy*t4/24.

        x[i] = ca*px - sa*py
        y[i] = ca*py + sa*px

    return x, y


@njit
def oblate_model_s(t, k, t0, p, a, aa, i, e, w, ldc,
                   mstar, rstar, ostar, tpole, gpole,
                   f, feff, sphi, cphi, beta, tref, fcoeffs,
                   tres, ts, xs, ys, xp, yp,
                   lcids, pbids, nsamples, exptimes, npb):
    y0, vx, vy, ax, ay, jx, jy, sx, sy = vajs_from_paiew(p, a, i, e, w)

    sa, ca = sin(aa), cos(aa)
    half_window_width = fmax(0.125, (2.0 + k[0])/vx)

    npt = t.size
    flux = zeros(npt)
    tp, lp = calculate_luminosity_interpolation_table(tres, k[0], xp, yp, sa, ca,
                                                      y0, vx, vy, ax, ay, jx, jy, sx, sy,
                                                      mstar, rstar, ostar, tpole, gpole, f, feff,
                                                      sphi, cphi, beta, ldc, tref, fcoeffs)
    dtp = tp[1] - tp[0]
    ls = create_star_luminosity(ts.size, xs, ys, mstar, rstar, ostar, tpole, gpole,
                                f, sphi, cphi, beta, ldc, tref, fcoeffs)

    astar = pi * (1. - feff)      # Area of an ellipse = pi * a * b, where a = 1 and b = (1 - feff)
    istar = zeros(npb)
    for ipb in range(npb):
        istar[ipb] = astar * nanmean(ls[ipb])

    for j in range(npt):
        epoch = floor((t[j] - t0 + 0.5*p)/p)
        tc = t[j] - (t0 + epoch*p)
        if abs(tc) > half_window_width:
            flux[j] = 1.0
        else:
            ilc = lcids[j]
            ipb = pbids[ilc]

            if k.size == 1:
                _k = k[0]
            else:
                _k = k[ipb]

            if isnan(_k) or isnan(a):
                flux[j] = inf
            else:
                for isample in range(1, nsamples[ilc] + 1):
                    time_offset = exptimes[ilc] * ((isample - 0.5) / nsamples[ilc] - 0.5)

                    to = tc + time_offset
                    it = int(floor((to - tp[0]) / dtp))
                    if it < 0:
                        it = 0
                        at = 0.0
                    elif it > lp.shape[1] - 2:
                        it = lp.shape[1] - 2
                        at = 1.0
                    else:
                        at = (to - tp[it]) / dtp
                    ml = (1.0 - at) * lp[ipb, it] + at * lp[ipb, it + 1]

                    x, y = xy_taylor_st(to, sa, ca, y0, vx, vy, ax, ay, jx, jy, sx, sy)

                    b = sqrt(x**2 + (y / (1. - feff))**2)
                    ia = circle_circle_intersection_area(1., _k, b)
                    flux[j] += (istar[ipb] - ml * ia) / istar[ipb]

                flux[j] /= nsamples[ilc]
    return flux


def map_osm(rstar, rho, rperiod, tpole, phi):
    omega = 2*pi/(rperiod*d2sec)  # Stellar rotation rate [rad/s]
    rho = 1e3*rho  # Stellar density       [kg/m^3]

    f = stellar_oblateness(omega, rho)  # Stellar oblateness
    feff = 1 - sqrt((1 - f)**2*cos(phi)**2 + sin(phi)**2)  # Projected stellar oblateness
    mstar = rho*4*pi/3*rstar**2*rstar*(1 - f)  # Stellar mass [kg]
    gpole = G*mstar/(rstar*(1 - f))**2  # Surface gravity at the pole  [m/s^2]
    return mstar, omega, gpole, f, feff


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2021  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import njit
from numpy import argsort, ceil, full, nan, zeros, sqrt, atleast_2d


@njit
def downsample_time_1d(time, vals, inttime=1., tmin=None, tmax=None, is_sorted=True):
    tmin = time.min() if tmin is None else tmin
    tmax = time.max() if tmax is None else tmax
    duration = tmax - tmin
    npt = time.size

    if not is_sorted:
        sids = argsort(time)
        time = time[sids]
        vals = vals[sids]

    nbins = int(ceil(duration / inttime))
    bt, bv, be = full(nbins, nan), zeros(nbins), zeros(nbins)

    ibin, tstart, istart, iend = 0, tmin, 0, 0
    while tstart <= tmax:
        while iend < npt and time[iend] <= tstart + inttime:
            iend += 1

        if iend > istart:
            bt[ibin] = time[istart:iend].mean()
            bv[ibin] = vals[istart:iend].mean()
            if iend - istart > 2:
                be[ibin] = vals[istart:iend].std() / sqrt(iend - istart)
            else:
                be[ibin] = nan

        istart = iend
        ibin += 1
        tstart += inttime
        if iend == npt:
            break
    return bt, bv, be


@njit
def downsample_time_2d(time, vals, inttime=1., tmin=None, tmax=None, is_sorted=True):
    vals = atleast_2d(vals)
    tmin = time.min() if tmin is None else tmin
    tmax = time.max() if tmax is None else tmax
    duration = tmax - tmin
    npt = time.size

    if not is_sorted:
        sids = argsort(time)
        time = time[sids]
        vals = vals[sids]

    nbins = int(ceil(duration / inttime))
    bt, bv, be = full(nbins, nan), zeros((nbins, vals.shape[1])), zeros((nbins, vals.shape[1]))

    ibin, tstart, istart, iend = 0, tmin, 0, 0
    while tstart <= tmax:
        while iend < npt and time[iend] <= tstart + inttime:
            iend += 1

        if iend > istart:
            nv = iend - istart
            bt[ibin] = time[istart:iend].mean()
            for j in range(vals.shape[1]):
                bv[ibin, j] = vals[istart:iend, j].mean()
            if nv > 2:
                for j in range(vals.shape[1]):
                    be[ibin, j] = vals[istart:iend, j].std() / sqrt(iend - istart)
            else:
                be[ibin, :] = nan

        istart = iend
        ibin += 1
        tstart += inttime
        if iend == npt:
            break
    return bt, bv, be



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pyopencl as cl
from numba import prange, njit
from numpy import atleast_2d, zeros, log, pi, asarray, unique, array, inf, where, isfinite, uint32, float64, int32, \
    arange

from ...param import  LParameter, NormalPrior as NP, UniformPrior as UP

@njit(parallel=False)
def psum2d(a):
    s = a.copy()
    ns = a.shape[0]
    for j in range(ns):
        midpoint = a.shape[1]
        while midpoint > 1:
            is_even = midpoint % 2 == 0
            if not is_even:
                s[j,0] += s[j,midpoint-1]
            midpoint = midpoint // 2
            for i in range(midpoint):
                s[j,i] = s[j,i] + s[j,midpoint+i]
    return s[:,0].astype(float64)


class CLLogLikelihood:
    def __init__(self, lpf, name: str = 'wn', noise_ids=None, sigma=None,
                 cl_ctx = None, cl_queue=None, cl_lnl_chunks: int = 1):
        self.name = name
        self.lpf = lpf
        self.cl_ctx = cl_ctx or self.lpf.tm.ctx
        self.cl_queue = cl_queue or self.lpf.tm.queue
        self.cl_lnl_chunks = cl_lnl_chunks

        # Define the OpenCL buffers
        # -------------------------
        self._b_flux = None
        self._b_err = None
        self._b_lnl2d = None
        self._b_lnl1d = None
        self._b_lcids = None
        self._b_errids = None

        if sigma is None:
            self.free = True
        else:
            self.sigma = asarray(sigma)
            self.free = False

        if noise_ids is not None:
            raise NotImplementedError("CLLoglikelihood cannot be applied to a subset of the data at the moment.")

        if lpf.noise_ids is None:
            raise ValueError('The LPF data needs to be initialised before initialising CLLogLikelihood.')

        self.global_noise_ids = noise_ids if noise_ids is not None else unique(lpf.noise_ids)
        self.mapping = {g:l for g,l in zip(self.global_noise_ids, arange(self.global_noise_ids.size))}

        slices, lnids = [], []
        for nid, sl in zip(lpf.noise_ids, lpf.lcslices):
            if nid in self.global_noise_ids:
                slices.append([sl.start, sl.stop])
                lnids.append(self.mapping[nid])
        self.lcslices = array(slices)
        self.local_pv_noise_ids = array(lnids, int32)

        self.times = lpf.timea
        self.fluxes = lpf.ofluxa

        if self.free:
            self.init_parameters()

        src = """
                   __kernel void lnl2d(const int nlc, __global const float *obs, __global const float *mod, __global const int *lcids,
                   __global const float *err, const int nerr, __global const int *errids, __global float *lnl2d){
                          uint i_tm = get_global_id(1);    // time vector index
                          uint n_tm = get_global_size(1);  // time vector size
                          uint i_pv = get_global_id(0);    // parameter vector index
                          uint n_pv = get_global_size(0);  // parameter vector population size
                          uint gid = i_pv*n_tm + i_tm;     // global linear index
                          uint ierr  = errids[lcids[i_tm]];
                          float e = err[i_pv*nerr + ierr];
                          lnl2d[gid] = -log(e) - 0.5f*log(2*M_PI_F) - 0.5f*pown((obs[i_tm]-mod[gid]) / e, 2);
                    }

                    __kernel void lnl1d(const uint npt, __global float *lnl2d, __global float *lnl1d){
                          uint i_pv = get_global_id(0);    // parameter vector index
                          uint n_pv = get_global_size(0);  // parameter vector population size

                        int i;
                        bool is_even;
                        uint midpoint = npt;
                        __global float *lnl = &lnl2d[i_pv*npt];

                        while(midpoint > 1){
                            is_even = midpoint % 2 == 0;   
                            if (is_even == 0){
                                lnl[0] += lnl[midpoint-1];
                            }
                            midpoint /= 2;

                            for(i=0; i<midpoint; i++){
                                lnl[i] = lnl[i] + lnl[midpoint+i];
                            }
                        }
                        lnl1d[i_pv] = lnl[0];
                    }

                    __kernel void lnl1d_chunked(const uint npt, __global float *lnl2d, __global float *lnl1d){
                        uint ipv = get_global_id(0);    // parameter vector index
                        uint npv = get_global_size(0);  // parameter vector population size
                        uint ibl = get_global_id(1);    // block index
                        uint nbl = get_global_size(1);  // number of blocks
                        uint lnp = npt / nbl;

                        __global float *lnl = &lnl2d[ipv*npt + ibl*lnp];

                        if(ibl == nbl-1){
                            lnp = npt - (ibl*lnp);
                        }

                        bool is_even;
                        uint midpoint = lnp;
                        while(midpoint > 1){
                            is_even = midpoint % 2 == 0;   
                            if (is_even == 0){
                                lnl[0] += lnl[midpoint-1];
                            }
                            midpoint /= 2;

                            for(int i=0; i<midpoint; i++){
                                lnl[i] = lnl[i] + lnl[midpoint+i];
                            }
                        }
                        lnl1d[ipv*nbl + ibl] = lnl[0];
                    }
                """
        self.prg_lnl = cl.Program(self.cl_ctx, src).build()
        self._init_data()

    def init_parameters(self):
        name = self.name
        pgp = [LParameter(f'{name}_loge_{i}', f'{name} log10 sigma {i}', '', UP(-4, 0), bounds=(-inf, inf)) for i in self.global_noise_ids]
        self.lpf.ps.thaw()
        self.lpf.ps.add_global_block(self.name, pgp)
        self.lpf.ps.freeze()
        self.pv_slice = self.lpf.ps.blocks[-1].slice
        self.pv_start = self.lpf.ps.blocks[-1].start
        setattr(self.lpf, f"_sl_{name}", self.pv_slice)
        setattr(self.lpf, f"_start_{name}", self.pv_start)

    def _init_data(self):
        lpf = self.lpf
        self.nlc = int32(lpf.nlc)
        self.n_noise_blocks = int32(lpf.n_noise_blocks)

        # Initialise the Python arrays
        # ----------------------------
        ninit = 1
        self.timea = lpf.timea.astype('f')
        self.ofluxa = lpf.ofluxa.astype('f')
        self.lnl2d = zeros([ninit, lpf.ofluxa.size], 'f')
        self.lnl1d = zeros([ninit, self.cl_lnl_chunks], 'f')
        self.ferr = zeros([ninit, lpf.nlc])
        self.lcids = lpf.lcids.astype('int32')
        self.pbids = lpf.pbids.astype('int32')
        self.noise_ids = lpf.noise_ids.astype('int32')

        # Release OpenCL buffers if they're initialised
        # ---------------------------------------------
        if self._b_flux:
            self._b_flux.release()
            self._b_err.release()
            self._b_lnl2d.release()
            self._b_lnl1d.release()
            self._b_lcids.release()
            self._b_errids.release()

        # Initialise OpenCL buffers
        # -------------------------
        mf = cl.mem_flags
        self._b_flux = cl.Buffer(self.cl_ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.ofluxa)
        self._b_err = cl.Buffer(self.cl_ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.ferr)
        self._b_lnl2d = cl.Buffer(self.cl_ctx, mf.WRITE_ONLY, self.lnl2d.nbytes)
        self._b_lnl1d = cl.Buffer(self.cl_ctx, mf.WRITE_ONLY, self.lnl1d.nbytes)
        self._b_lcids = cl.Buffer(self.cl_ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.lcids)
        self._b_errids = cl.Buffer(self.cl_ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.local_pv_noise_ids)

    def _lnl2d(self, pv):
        if self.lnl2d.shape[0] != pv.shape[0] or self.lnl1d.size != pv.shape[0]:
            self.err = zeros([pv.shape[0], self.n_noise_blocks], 'f')
            self._b_err.release()
            self._b_err = cl.Buffer(self.cl_ctx, cl.mem_flags.WRITE_ONLY, self.err.nbytes)
            self.lnl2d = zeros([pv.shape[0], self.ofluxa.size], 'f')
            self._b_lnl2d.release()
            self._b_lnl2d = cl.Buffer(self.cl_ctx, cl.mem_flags.WRITE_ONLY, self.lnl2d.nbytes)
            self.lnl1d = zeros([pv.shape[0], self.cl_lnl_chunks], 'f')
            if self._b_lnl1d:
                self._b_lnl1d.release()
            self._b_lnl1d = cl.Buffer(self.cl_ctx, cl.mem_flags.WRITE_ONLY, self.lnl1d.nbytes)

        err = (10 ** pv[:, self.pv_slice]).astype('f')
        cl.enqueue_copy(self.cl_queue, self._b_err, err)
        self.prg_lnl.lnl2d(self.cl_queue, self.lpf.tm.f.shape, None, self.nlc, self._b_flux, self.lpf.tm._b_f,
                           self._b_lcids, self._b_err, self.n_noise_blocks, self._b_errids, self._b_lnl2d)

    def lnlikelihood_numba(self, pv):
        self._lnl2d(pv)
        cl.enqueue_copy(self.cl_queue, self.lnl2d, self._b_lnl2d)
        lnl = psum2d(self.lnl2d)
        return where(isfinite(lnl), lnl, -inf)

    def lnlikelihood_ocl(self, pv):
        self._lnl2d(pv)
        self.prg_lnl.lnl1d_chunked(self.cl_queue, [self.lnl2d.shape[0], self.cl_lnl_chunks], None,
                                   uint32(self.lnl2d.shape[1]), self._b_lnl2d, self._b_lnl1d)
        cl.enqueue_copy(self.cl_queue, self.lnl1d, self._b_lnl1d)
        lnl = self.lnl1d.astype('d').sum(1)
        return lnl

    def lnlikelihood_numpy(self, pv):
        self._lnl2d(pv)
        cl.enqueue_copy(self.cl_queue, self.lnl2d, self._b_lnl2d)
        lnl = self.lnl2d.astype('d').sum(1)
        return where(isfinite(lnl), lnl, -inf)

    def __call__(self, pvp, model):
        return self.lnlikelihood_ocl(pvp)



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from numpy import inf, isfinite, abs, sum, unique, ones, compress, array, mean, round, int, where, float64, full_like, \
    nan

from pytransit.orbits import epoch


def fold_orbit_and_phase(time, period, origo, shift):
    phase  = ((time - origo) / period + shift)
    orbit  = (phase // 1).astype(np.int)
    phase  = (phase % 1. - 0.5) * period
    return  orbit, phase

class KeplerLC(object):
    def __init__(self, time, flux, quarter, zero_epoch, period, d_transit, d_baseline, error=None, **kwargs):
        """
        Kepler light curve.

        A convenience class to help working with Kepler light curves.

        Parameters
        ----------

          time       : BJD array      [d]
          flux       : Flux array     [-]
          zero_epoch : zero epoch     [d]
          period     : orbital period [d]
          d_transit  : transit duration (approximate) [d]
          d_baseline : total duration to include around each transit
          quarter    : quarter array

          max_ptp    : maximum point-to-point span per transit
          min_pts    : minimum number of points per transit
        """
        self.t0   = zero_epoch
        self.p    = period
        self.max_ptp = kwargs.get('max_ptp', inf)
        self.min_pts = kwargs.get('min_pts', 5)

        orbit, phase = fold_orbit_and_phase(time, period, zero_epoch, 0.5)
        orbit -= orbit.min()
        msk_phase = abs(phase) < 0.5*d_baseline             # phase inclusion mask
        msk_oot   = abs(phase) > 0.5*d_transit              # out-of-transit mask
        msk_inc   = isfinite(time) & isfinite(flux) & msk_phase        # final data inclusion mask

        self.time   = array(time, float64)
        self.phase = phase
        self.flux   = array(flux, float64)
        self.qidarr = array(quarter)                                   # quarter indices
        self.tidarr, nt = orbit, orbit[-1]                             # transit indices
        self.msk_oot = msk_oot
        self.error = error if error is not None else full_like(self.time, nan)

        # Remove orbits with too few datapoints
        # -------------------------------------
        npts = array([sum((self.tidarr == tid) & msk_inc) for tid in unique(self.tidarr)])
        for tid,npt in enumerate(npts):
            if npt < self.min_pts:
                msk_inc[self.tidarr==tid] = 0
        self._compress_data(msk_inc)
        self._compute_indices()

        # Remove orbits with too big ptp range
        # ------------------------------------
        msk_inc = ones(self.npt, np.bool)
        for tid,ptp in enumerate(list(map(np.ptp, self.normalized_flux_per_transit))):
            if ptp > self.max_ptp:
                msk_inc[self.tidarr==tid] = 0
        self._compress_data(msk_inc)
        self._compute_indices()

    def _compress_data(self, mask):
        self.time    = compress(mask, self.time)
        self.phase   = compress(mask, self.phase)
        self.flux    = compress(mask, self.flux)
        self.error   = compress(mask, self.error)
        self.qidarr  = compress(mask, self.qidarr)
        self.tidarr  = compress(mask, self.tidarr)
        self.msk_oot = compress(mask, self.msk_oot)

    def _compute_indices(self): 
        self.qids = unique(self.qidarr)
        self.tids = unique(self.tidarr)
        self.nt   = len(self.tids)
        self.npt  = self.time.size

        self.qslices = [slice(*where(self.qidarr==qid)[0][[0,-1]]+[0,1]) for qid in self.qids]
        self.qsldict = {qid : slice(*where(self.qidarr==qid)[0][[0,-1]]+[0,1]) for qid in self.qids}

        for i,tid in enumerate(self.tids):
            self.tidarr[self.tidarr==tid] = i
        self.tids = unique(self.tidarr)
        self.tslices = [slice(*where(self.tidarr==tid)[0][[0,-1]]+[0,1]) for tid in self.tids]
        self.orbit_n = array([epoch(t.mean(), self.t0, self.p) for t in self.time_per_transit])

    def get_transit(self, tid, normalize=False, mask_transit=False):
        mask = self.tidarr==tid
        return self.time[mask], self.normalized_flux[mask] if normalize else self.flux[mask]

    def remove_common_orbits(self, lc2):
        is_unique = ~np.in1d(self.orbit_n, lc2.orbit_n)
        mask = np.ones(self.npt, np.bool)
        for tid,include in enumerate(is_unique):
            mask[self.tslices[tid]] = include
        self._compress_data(mask)
        self._compute_indices()

    @property
    def flux_per_transit(self):
        return [self.flux[sl] for sl in self.tslices]
    
    @property
    def normalized_flux_per_transit(self):
        nf = self.normalized_flux
        return [nf[sl] for sl in self.tslices]

    @property
    def normalized_oot_flux_per_transit(self):
        nf = self.normalized_flux
        return [nf[sl][self.oot_mask[sl]] for sl in self.tslices]

    @property
    def time_per_transit(self):
        return [self.time[sl] for sl in self.tslices]

    @property
    def oot_time_per_transit(self):
        return [self.time[sl][self.oot_mask[sl]] for sl in self.tslices]

    @property
    def error_per_transit(self):
        return [self.error[sl] for sl in self.tslices]

    @property
    def normalized_error(self):
        return self.error / self.flux_baseline

    @property
    def normalized_error_per_transit(self):
        nf = self.normalized_error
        return [nf[sl] for sl in self.tslices]

    @property
    def quarter_per_transit(self):
        return [self.qidarr[sl] for sl in self.tslices]

    @property
    def oot_mask(self):
        return self.msk_oot

    @property
    def normalized_flux(self):
        return self.flux / self.flux_baseline

    @property
    def flux_baseline(self):
        bl = np.zeros_like(self.flux)
        for tid in self.tids:
            mask_tid = tid==self.tidarr
            mask_nrm = mask_tid & self.msk_oot
            bl[mask_tid] = np.median(self.flux[mask_nrm])
        return bl

    @property
    def oot_flux_std(self):
        return self.normalized_flux[self.msk_oot].std()



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2021  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numpy import inf, repeat, atleast_2d, sqrt, arange

from ..lpf import map_ldc
from ..tesslpf import TESSLPF
from ...orbits import as_from_rhop, i_from_ba
from ...param import GParameter, NormalPrior as NP, UniformPrior as UP


class TESSTTVLPF(TESSLPF):

    def _init_p_orbit(self):
        """Orbit parameter initialisation.
        """
        porbit = [
            GParameter('p', 'period', 'd', NP(1.0, 1e-5), (0, inf)),
            GParameter('rho', 'stellar_density', 'g/cm^3', UP(0.1, 25.0), (0, inf)),
            GParameter('b', 'impact_parameter', 'R_s', UP(0.0, 1.0), (0, 1))]
        self.ps.add_global_block('orbit', porbit)

        ptc = [GParameter(f'tc_{i}', f'transit_center_{i}', '-', NP(0.0, 0.1), (-inf, inf)) for i in range(self.nlc)]
        self.ps.add_global_block('tc', ptc)
        self._pid_tc = repeat(self.ps.blocks[-1].start, self.nlc)
        self._start_tc = self.ps.blocks[-1].start
        self._sl_tc = self.ps.blocks[-1].slice

    def transit_model(self, pv, copy=True):
        pv = atleast_2d(pv)
        ldc = map_ldc(pv[:, self._sl_ld])
        zero_epoch = pv[:, self._sl_tc] - self._tref
        period = pv[:, 0]
        smaxis = as_from_rhop(pv[:, 1], period)
        inclination = i_from_ba(pv[:, 2], smaxis)
        radius_ratio = sqrt(pv[:, self._sl_k2])
        return self.tm.evaluate(radius_ratio, ldc, zero_epoch, period, smaxis, inclination)

    def _post_initialisation(self):
        super()._post_initialisation()
        self.tm.epids = arange(self.nlc)


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from pathlib import Path
from typing import Union

import pandas as pd
import astropy.units as u
from numba import njit
from numpy import ndarray, pi, atleast_1d, zeros, exp, diff, log, sqrt, nan, vstack, tile, linspace, cos, sin, sum, \
    isfinite
from scipy.constants import c,h,k,G
from scipy.optimize import brentq

from ..stars import create_husser2013_interpolator, create_bt_settl_interpolator
#from ..contamination.filter import Filter

NPType = Union[float,ndarray]

mj2kg = u.M_jup.to(u.kg)
ms2kg = u.M_sun.to(u.kg)
d2s = u.day.to(u.s)

def equilibrium_temperature(tstar: NPType, a: NPType, f: NPType, ab: NPType) -> NPType:
    """Planetary equilibrium temperature [K].

    Parameters
    ----------
    tstar
        Effective stellar temperature  [K]
    a
        Scaled semi-major axis [Rsun]
    f
        Redistribution factor
    ab
        Bond albedo

    Returns
    -------
    Teq : float or ndarray
        Equilibrium temperature [K]
    """
    return tstar * sqrt(1 / a) * (f * (1 - ab)) ** 0.25


# Thermal emission
# ================

def planck(t: NPType, l: NPType) -> NPType:
    """Radiance of a black body as a function of wavelength.

    Parameters
    ----------
    t
        Black body temperature [K]
    l
        Wavelength [m]

    Returns
    -------
    L  : float or ndarray
        Back body radiance [W m^-2 sr^-1]
    """
    return 2 * h * c ** 2 / l ** 5 / (exp(h * c / (l * k * t)) - 1)

@njit
def summed_planck(teff, wl, tm):
    teff = atleast_1d(teff)
    flux = zeros(teff.shape[0])
    for i in range(flux.size):
        flux[i] = sum(tm*(2*h*c**2 / wl**5 / (exp(h*c / (wl*k*teff[i])) - 1.)))
    return flux


def emission(tp: NPType, tstar: NPType, k: NPType, flt) -> NPType:
    """Thermal emission from the planet.

    Parameters
    ----------
    tp : float or ndarray
        Equilibrium temperature of the planet [K]
    tstar : float or ndarray
        Effective temperature of the star [K]
    k : float or ndarray
        Planet-star radius ratio
    flt: Filter
        Passband transmission

    Returns
    -------
    float or ndarray

    References
    """
    wl = linspace(flt.wl_min, flt.wl_max, 100)
    tm = flt(wl)
    return k**2 * (summed_planck(tp, wl, tm) / summed_planck(tstar, wl, tm))


# Doppler boosting
# ================

def doppler_beaming_factor(teff: float, flt, dataset: str = 'BT-Settl'):
    """The photon weighted bandpass-integrated Doppler beaming factor.

    Parameters
    ----------
    teff
        Effective temperature of the star [K]
    flt
        Passband transmission
    dataset
        Either 'BT-Settl' or 'Husser2013'

    Returns
    -------
    float
        The photon weighted bandpass-integrated Doppler beaming factor.

    Notes
    ------
    See Bloemen et al., MNRAS, 410 (2010) and Claret, A. et al., A&A, 641 (2020).
    """
    if dataset == 'Husser2013':
        ip = create_husser2013_interpolator()
    else:
        ip = create_bt_settl_interpolator()

    wl_nm = ip.grid[1][1:-1]
    wl = 1e-9 * wl_nm
    fl = ip(vstack((tile(teff, wl_nm.size), wl_nm)).T)
    b = 5. + diff(log(fl)) / diff(log(wl))
    w1 = flt(wl_nm)[1:] * wl[1:] * fl[1:]
    w2 = flt(wl_nm)[:-1] * wl[:-1] * fl[:-1]
    w = 0.5 * (w1 + w2)
    m = isfinite(b)
    return sum(w[m] * b[m]) / sum(w[m])


def doppler_beaming_amplitude(mp: NPType, ms: NPType, period: NPType, b: NPType) -> NPType:
    """The amplitude of the Doppler beaming signal.

    Calculates the amplitude of the Doppler beaming signal following the approach
    described by Loeb & Gaudi in [Loeb2003]_ . Note that you need to pre-calculate the photon-weighted bandpass-integrated
    doppler Beaming factor [Bloemen2010]_ [Barclay2012]_ for the star and the instrument using ``doppler_beaming_factor``.

    Parameters
    ----------
    mp : float or ndarray
        Planetary mass [MJup]
    ms : float or ndarray
        Stellar mass [MSun]
    period : float or ndarray
        Orbital period [d]
    b : float or ndarray
        Photon-weighted bandpass-integrated Doppler beaming factor

    Returns
    -------
    float or ndarray
        Doppler boosting signal amplitude

    References
    ----------
    .. [Loeb2003] Loeb, A. & Gaudi, B. S. Periodic Flux Variability of Stars due to the Reflex Doppler Effect Induced by
           Planetary Companions. Astrophys. J. 588, L117–L120 (2003).
    .. [Bloemen2010] Bloemen, S. et al. Kepler observations of the beaming binary KPD 1946+4340. MNRAS 410, (2010).
    .. [Barclay2012] Barclay, T. et al. PHOTOMETRICALLY DERIVED MASSES AND RADII OF THE PLANET AND STAR IN THE TrES-2
           SYSTEM. AspJ 761, 53 (2012).
    """
    return b / c *(2*pi*G/(d2s*period))**(1/3) * ((mp*mj2kg)/(ms*ms2kg)**(2/3))

# Ellipsoidal variations
# ======================

def ellipsoidal_variation(f: NPType, theta: NPType, mp: float, ms: float, a: float, i: float, e: float, u: float, g: float):
    raise NotImplementedError
    #return ellipsoidal_variation_amplitude(mp, ms, a, i, u, g)

def ellipsoidal_variation_signal(f: NPType, theta: NPType, e: float) -> NPType:
    """

    Parameters
    ----------
    f
        True anomaly [rad]
    theta
        Angle between the line-of-sight and the star-planet direction
    e
        Eccentricity

    Returns
    -------

    """
    return -((1 + e * cos(f)) / (1-e**2))**3 * cos(2*theta)


def ellipsoidal_variation_amplitude(mp: NPType, ms: NPType, a: NPType, i: NPType, u: NPType, g: NPType) -> NPType:
    """The amplitude of the ellipsoidal variation signal.

    Calculates the amplitude of the ellipsoidal variation signal following the approach described by
    Lillo-Box et al. in [Lillo-Box2014]_, page 11.

    Parameters
    ----------
    mp : float or ndarray
        Planetary mass [MJup]
    ms : float or ndarray
        Stellar mass [MSun]
    a : float or ndarray
        Semi-major axis of the orbit divided by the stellar radius
    i : float or ndarray
        Orbital inclination [rad]
    u : float or ndarray
        Linear limb darkening coefficient
    g : float or ndarray
        Gravity darkening coefficient

    Returns
    -------
    ev_amplitude: float or ndarray
        The amplitude of the ellipsoidal variation signal

    References
    ----------
    .. [Lillo-Box2014] Lillo-Box, J. et al. Kepler-91b: a planet at the end of its life. A&A 562, A109 (2014).

    """
    ae = 0.15 * (15 + u) * (1 + g) / (3 - g)
    return ae * (mp*mj2kg)/(ms*ms2kg) * a**-3 * sin(i)**2


def reflected_fr(a: NPType, ab: NPType, r: NPType = 1.5) -> NPType:
    """Reflected flux ratio per projected area element.

    Parameters
    ----------
    a
        Scaled semi-major axis [Rsun]
    ab
        Bond albedo
    r
        Inverse of the phase integral

    Returns
    -------
    fr : float
        Reflected flux ratio
    """
    return r * ab / a ** 2

def flux_ratio(tstar: NPType, a: NPType, f: NPType, ab: NPType, l: NPType, r: NPType = 1.5, ti: NPType = 0) -> NPType:
    """Total flux ratio per projected area element.

    Parameters
    ----------
    tstar
        Effective stellar temperature [K]
    a
        Scaled semi-major axis [Rs]
    f
        Redistribution factor
    ab
        Bond albedo
    l
        Wavelength [m]
    r
        Inverse of the phase integral
    ti
        Temperature [K]

    Returns
    -------
    fr: float
        Total flux ratio
    """
    return reflected_fr(a, ab, r) + thermal_fr(tstar, a, f, ab, l, ti)


def solve_teq(fr, tstar, a, ab, l, r=1.5, ti=0):
    """Solve the equilibrium temperature.

    Parameters
    ----------
    fr
        Flux ratio
    tstar
        Effective stellar temperature [K]
    a
        Scaled semi-major axis [Rs]
    ab
        Bond albedo
    l
        Wavelength [m]
    r
        Inverse of the phase integral
    ti
        Temperature [K]

    Returns
    -------
    Teq : float or ndarray
        Equilibrium temperature

    """
    Bs = planck(tstar, l)
    try:
        return brentq(lambda Teq: reflected_fr(a, ab, r) + planck(Teq + ti, l) / Bs - fr, 5, tstar)
    except ValueError:
        return nan


def solve_ab(fr, tstar, a, f, l, r=1.5, ti=0):
    """Solve the Bond albedo.

    Parameters
    ----------
    fr  :
        Flux ratio                    [-]
    tstar  :
        Effective stellar temperature [K]
    a   :
        Scaled semi-major axis        [Rs]
    A   :
        Bond albedo                   [-]
    l   :
        Wavelength                    [m]
    r   :
        Inverse of the phase integral [-]
    ti  :
        Temperature                   [K]

    Returns
    -------
      A : Bond albedo
    """
    try:
        return brentq(lambda ab: reflected_fr(a, ab, r) + thermal_fr(tstar, a, f, ab, l, ti) - fr, 0, 0.3)
    except ValueError:
        return nan


def solve_redistribution(fr, tstar, a, ab, l):
    """Solve the redistribution factor.

    Parameters
    ----------
    fr  :
        Flux ratio                    [-]
    tstar  :
        Effective stellar temperature [K]
    a   :
        Scaled semi-major axis        [Rs]
    ab   :
        Bond albedo                   [-]
    l   :
        Wavelength                    [m]
    r   :
        Inverse of the phase integral [-]
    Ti  : t
        Temperature                   [K]

    Returns
    -------
      f : Redistribution factor
    """
    Teqs = solve_teq(fr, tstar, l)
    return brentq(lambda f: equilibrium_temperature(tstar, a, f, ab) - Teqs, 0.25, 15)



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math as m

from numpy import inf, zeros, pi, log, where, exp
from numpy.random import normal, uniform
from scipy.stats import gamma as gm, laplace


class Prior:
    def __init__(self):
        raise NotImplementedError

    def logpdf(self, v):
        raise NotImplementedError

    def rvs(self, size):
        raise NotImplementedError


class DefaultPrior(Prior):
    def logpdf(self, v: float):
        return 0

    def rvs(self, size):
        return zeros(size)


class NormalPrior(Prior):
    def __init__(self, mean: float, std: float):
        self.mean = float(mean)
        self.std = float(std)
        self._f1 = 1 / m.sqrt(2*pi*std**2)
        self._lf1 = m.log(self._f1)
        self._f2 = 1 / (2*std**2)

    def logpdf(self, x):
        return self._lf1 - self._f2*(x - self.mean)**2

    def rvs(self, size=1):
        return normal(self.mean, self.std, size)

    def __str__(self):
        return f'N(μ = {self.mean}, σ = {self.std})'

    def __repr__(self):
        return f'NormalPrior({self.mean}, {self.std})'


class UniformPrior(Prior):
    def __init__(self, a: float, b: float):
        self.a, self.b = a, b
        self.lnc = m.log(b-a)

    def logpdf(self, v):
        return where((self.a < v) & (v < self.b), self.lnc, -inf)

    def rvs(self, size=1):
        return uniform(self.a, self.b, size)

    def __str__(self):
        return f'U(a = {self.a}, b = {self.b})'

    def __repr__(self):
        return f'UniformPrior({self.a}, {self.b})'


class JeffreysPrior(Prior):
    def __init__(self, x0: float, x1: float):
        self.x0 = x0
        self.x1 = x1
        self._f = log(x1 / x0)

    def pdf(self, x):
        return where((x > self.x0) & (x < self.x1), 1. / (x * self._f), -inf)

    def logpdf(self, x):
        return where((x > self.x0) & (x < self.x1), -log(x * self._f), -inf)

    def rvs(self, size=1):
        return exp(uniform(log(self.x0), log(self.x1), size))


class LaplacePrior(Prior):
    def __init__(self, mean, mad):
        self.mean = mean
        self.mad = mad
        self._n = 1. / (2 * mad)
        self._ln = log(self._n)
        self._p = laplace(mean, mad)

    def logpdf(self, v):
        return self._ln - abs(v - self.mean) / self.mad

    def rvs(self, size):
        return self._p.rvs(size)

    def __str__(self):
        return f'L(μ = {self.mean}, MAD = {self.mad})'

    def __repr__(self):
        return f'LaplacePrior({self.mean}, {self.mad})'


class LogLogisticPrior(Prior):
    def __init__(self, a, b):
        self.a, self.b = a, b

    def logpdf(self, v):
        if not 1e-3 < v < 1.:
            return -inf
        else:
            a, b = self.a, self.b
            return m.log((b / a) * (v / a) ** (b - 1.) / (1. + (v / a) ** b) ** 2)

    def rvs(self, size=1):
        return uniform(1e-3, 1.0, size)


class GammaPrior(Prior):
    def __init__(self, a):
        self.a = a
        self.A = -m.lgamma(a)

    def logpdf(self, x):
        return self.A + (self.a - 1.) * log(x) - x

    def rvs(self, size):
        return gm(self.a).rvs(size)



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numpy import asarray, unique, zeros, inf, squeeze, zeros_like, isfinite, log10, diff, sqrt, arange, log
from scipy.optimize import minimize
from tqdm.auto import tqdm

try:
    from celerite import GP
    from celerite.terms import Matern32Term
    with_celerite = True
except ImportError:
    with_celerite = False

class FrozenMultiCeleriteLogLikelihood:
    def __init__(self, lpf, name: str = 'gp', lcids=None):
        if not with_celerite:
            raise ImportError("CeleriteLogLikelihood requires celerite.")

        self.name = name
        self.lpf = lpf

        if lpf.lcids is None:
            raise ValueError('The LPF data needs to be initialised before initialising CeleriteLogLikelihood.')
        self.lcids = lcids if lcids is not None else arange(lpf.nlc)
        self.gps = []

        self.learn_gps()

    def learn_gps(self):
        for lcid in tqdm(self.lcids, desc='Learning GPs', leave=False):
            kernel = Matern32Term(log(self.lpf.fluxes[lcid].std()), log(0.1))
            gp = GP(kernel, mean=0.0)
            gp.freeze_parameter('kernel:log_sigma')

            def nll(x):
                gp.set_parameter_vector(x)
                gp.compute(self.lpf.times[lcid], self.lpf.wn[lcid])
                return -gp.log_likelihood(self.lpf.fluxes[lcid] - 1.0)

            res = minimize(nll, [log(1.)])
            gp.set_parameter_vector(res.x)
            gp.compute(self.lpf.times[lcid], self.lpf.wn[lcid])
            self.gps.append(gp)

    def compute_gp_lnlikelihood(self, pv, model):
        lnl = 0.0
        for lcid, gp in zip(self.lcids, self.gps):
            sl = self.lpf.lcslices[lcid]
            lnl += gp.log_likelihood(self.lpf.fluxes[lcid] - model[sl])
        return lnl

    def predict_baseline(self, pv):
        residuals = self.lpf.ofluxa - squeeze(self.lpf.transit_model(pv))
        bl = zeros_like(self.lpf.timea)
        for lcid, gp in zip(self.lcids, self.gps):
            sl = self.lpf.lcslices[lcid]
            bl[sl] = gp.predict(residuals[sl], self.lpf.times[lcid], return_cov=False)
        return 1. + bl

    def __call__(self, pvp, model):
        if pvp.ndim == 1:
            lnlike = self.compute_gp_lnlikelihood(pvp, model)
        else:
            lnlike = zeros(pvp.shape[0])
            for ipv, pv in enumerate(pvp):
                if all(isfinite(model[ipv])):
                    lnlike[ipv] = self.compute_gp_lnlikelihood(pv, model[ipv])
                else:
                    lnlike[ipv] = -inf
        return lnlike


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numpy import pi, array
from .orbits_py import epoch, fold, mean_anomaly, duration_eccentric, eclipse_phase, ea_iter_v, ea_newton_v, z_ps5, z_newton_v, \
    z_ps3, z_iter_v, ta_ip, ta_ps5, ta_ps3, ta_newton_v, ta_iter_v, z_circular

TWO_PI = 2 * pi


def extract_time_transit(time, k, tc, p, a, i, e, w, td_factor=1.1):
    """ Extracts the in-transit times from a light curve.

    Parameters
    ----------
    time
    k
    tc
    p
    a
    i
    e
    w
    td_factor

    Returns
    -------

    """
    td = duration_eccentric(p, k, a, i, e, w, 1)
    folded_time = (time - tc + 0.5*p) % p - 0.5*p
    mask = abs(folded_time) < td_factor*0.5*td
    return time[mask], mask


def extract_time_eclipse(time, k, tc, p, a, i, e, w, td_factor=1.1):
    """ Extracts the in-eclipse times from a light curve.

    Parameters
    ----------
    time
    k
    tc
    p
    a
    i
    e
    w
    td_factor

    Returns
    -------

    """
    td  = duration_eccentric(p, k, a, i, e, w, 1)
    tc += eclipse_phase(p, i, e, w)
    folded_time = (time - tc + 0.5*p) % p - 0.5*p
    mask = abs(folded_time) < td_factor*0.5*td
    return time[mask], mask


def not_implemented(*nargs, **kwargs):
    raise NotImplementedError


class Orbit(object):
    methods = 'iteration newton ps3 ps5 interpolation'.split()

    ea_functions = dict(iteration=ea_iter_v,
                        newton=ea_newton_v,
                        ps3=not_implemented,
                        ps5=not_implemented,
                        interpolation=not_implemented)

    ta_functions = dict(iteration=ta_iter_v,
                        newton=ta_newton_v,
                        ps3=ta_ps3,
                        ps5=ta_ps5,
                        interpolation=ta_ip)

    z_functions = dict(iteration=z_iter_v,
                       newton=z_newton_v,
                       ps3=z_ps3,
                       ps5=z_ps5,
                       interpolation=z_iter_v)

    def __init__(self, method='iteration', circular_e_threshold=1e-2):
        assert method in self.methods
        self.method = method
        self._mine = circular_e_threshold
        self._ea_function = self.ea_functions[method]
        self._ta_function = self.ta_functions[method]
        self._z_function = self.z_functions[method]

    def mean_anomaly(self, time, t0, p, e=0., w=0.):
        return mean_anomaly(time, t0, p, e, w)

    def eccentric_anomaly(self, time, t0, p, e=0., w=0.):
        if e > self._mine:
            return self._ea_function(time, t0, p, e, w)
        else:
            return self.mean_anomaly(time, t0, p)

    def true_anomaly(self, time, t0, p, e=0., w=0.):
        if e > self._mine:
            return self._ta_function(time, t0, p, e, w)
        else:
            return self.mean_anomaly(time, t0, p)

    def projected_distance(self, time, t0, p, a, i, e=0., w=0.):
        if e > self._mine:
            return self._z_function(time, array([t0, p, a, i, e, w]))
        else:
            return z_circular(time, array([t0, p, a, i, e, w]))

    def phase(self, time, t0, p, a, i, e=0., w=0.):
        raise NotImplementedError


class CircularOrbit(Orbit):
    def mean_anomaly(self, time, t0, p):
        return ((time - t0) / p) * TWO_PI

    def eccentric_anomaly(self, time, t0, p):
        return self.mean_anomaly(time, t0, p)

    def true_anomaly(self, time, t0, p):
        return self.mean_anomaly(time, t0, p)

    def projected_distance(self, time, t0, p, a, i):
        return z_circular(time, t0, p, a, i)



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from ..orbits.orbits_py import a_from_mp
from .constants import msun, mjup

as_c = {'as':1., 'mas':1e3, 'muas':1e6}

def angular_signal(mass, period, distance, mstar=1., units='as'):
    """Angular semi-amplitude of the astrometric signal caused by a planet.
    
    Parameters
    ----------

      mass     : planet mass            [M_Jup]
      period   : orbital period         [d]
      distance : distance to the system [pc]
      mstar    : stellar mass           [M_Sun]
      units    : 'as', 'mas', or 'muas'

    Returns
    -------

      angular semi-amplitude of the astrometric signal by the planet in given units

    """
    return as_c[units] * (mass*mjup)/(mstar*msun) * a_from_mp(mstar, period) / distance



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import pyopencl as cl
from numba import njit, float64
from numpy import inf, sqrt, zeros, isfinite, where, atleast_2d, int32, uint32

from .loglikelihood.clloglikelihood import CLLogLikelihood
from ..models.ma_quadratic_cl import QuadraticModelCL
from ..orbits.orbits_py import as_from_rhop, i_from_ba
from .lpf import BaseLPF

class OCLBaseLPF(BaseLPF):
    def __init__(self, target: str, passbands: list, times: list = None, fluxes: list = None, errors: list = None,
                 pbids: list = None, covariates: list = None, nsamples: tuple = None, exptimes: tuple = None,
                 wnids: list = None,
                 klims: tuple = (0.01, 0.75), nk: int = 512, nz: int = 512, cl_ctx=None, cl_queue=None, init_data=True,
                 cl_lnl_chunks: int = 1):

        self.cl_ctx = cl_ctx or self.tm.ctx
        self.cl_queue = cl_queue or self.tm.queue
        self.cl_lnl_chunks = cl_lnl_chunks

        tm = QuadraticModelCL(klims=klims, nk=nk, nz=nz, cl_ctx=cl_ctx, cl_queue=cl_queue)
        super().__init__(target, passbands, times, fluxes, errors, pbids, covariates, wnids, tm, nsamples, exptimes,
                         init_data=init_data)

    def _init_lnlikelihood(self):
        self._add_lnlikelihood_model(CLLogLikelihood(self, cl_lnl_chunks=self.cl_lnl_chunks))

    def transit_model(self, pvp, copy=False):
        pvp = atleast_2d(pvp)
        pvp_t = zeros([pvp.shape[0], 8], "f")
        uv = zeros([pvp.shape[0], 2], "f")
        pvp_t[:, 0:1] = sqrt(pvp[:, self._pid_k2])  # Radius ratio
        pvp_t[:, 1:3] = pvp[:, 0:2]                 # Transit centre and orbital period
        pvp_t[:, 3] = a = as_from_rhop(pvp[:, 2], pvp[:, 1])
        pvp_t[:, 4] = i_from_ba(pvp[:, 3], a)
        a, b = sqrt(pvp[:, self._sl_ld][:, 0]), 2. * pvp[:, self._sl_ld][:, 1]
        uv[:, 0] = a * b
        uv[:, 1] = a * (1. - b)
        flux = self.tm.evaluate_pv(pvp_t, uv, copy=copy)
        return flux if copy else None

    def flux_model(self, pvp):
        return self.transit_model(pvp, copy=True).astype('d')

    def lnlikelihood(self, pvp):
        """Log likelihood for a 1D or 2D array of model parameters.

        Parameters
        ----------
        pvp: ndarray
            Either a 1D parameter vector or a 2D parameter array.

        Returns
        -------
            Log likelihood for the given parameter vector(s).
        """
        pvp = atleast_2d(pvp)
        self.transit_model(pvp, copy=False)
        return self._lnlikelihood_models[0](pvp)



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple
from numpy import linspace, trapz, sqrt, ndarray, zeros, pi


class LDModel:
    def __init__(self, niz: int = 200):
        self._int_z = linspace(0, 1, niz)
        self._int_mu = sqrt(1 - self._int_z ** 2)

    def __call__(self, mu: ndarray, x: ndarray) -> Tuple[ndarray, ndarray]:
        return self._evaluate(mu, x), self._integrate(x)

    def _evaluate(self, mu: ndarray, x:ndarray) -> ndarray:
        raise NotImplementedError

    def _integrate(self, x: ndarray) -> ndarray:
        npv = x.shape[0]
        npb = x.shape[1]
        ldi = zeros((npv, npb))
        for ipv in range(npv):
            for ipb in range(npb):
                ldi[ipv,ipb] = 2. * pi * trapz(self._int_z * self(self._int_mu, x), self._int_z)
        return ldi